/*****************************************************************************/
/*                                                                           */
/*  HyperPanel OS                       File "stm32wl/src/signal_vmit.h"     */
/*  Board - Constants and definitions specific to stm32wl.                   */
/*                                                                           */
/*  Author : TE                         Last modification : 06/10/2023       */
/*                                                                           */
/*****************************************************************************/

#ifndef SIGNAL_H                       /* Begin conditionnal compilation     */
#define SIGNAL_H                       /* Identifier for this include        */


/* Macro for "vmit.c" .......................................................*/

#include <armv7m.h>
extern const Ref_nvic ref_nvic[]  ;

#define SIGNAL_RAISE(a) {register INT nirq                     ; \
                         nirq = ref_nvic[a].num                ; \
                         nvic->isp[nirq/32] = 1 << (nirq % 32) ; }

#define SIGNAL_CLEAR(a) {register INT nirq                     ; \
                         nirq = ref_nvic[a].num                ; \
                         nvic->icp[nirq/32] = 1 << (nirq % 32) ; }


/* Bench app - DWT samplig macro for tl_vmitth.c ............................*/

#define DWT_CYCC (*(UINT*) 0xE0001004) /* Not used on stm32wl.               */
#define ADD_CYCC_IRQ                   /* Not used on stm32wl.               */

/* Bench app - GPIO macro for oscilloscope profiling ..........................

   Name    ST name  Function              Macro           
   ------  -------  --------------------  -----------------
   GPIO1   PE0      BOOT                  GPIO_BOOT(outin)
   GPIO2   PE1      HRT synchro input     
   GPIO3   PE2      SysTick output        GPIO_TIM (outin)
   GPIO4   PE3      PCM output signal     

   GPIO5   PE4      CPU                   GPIO_CPU(outin)
   GPIO6   PE5      VMIT                  GPIO_VMIT(outin)
   GPIO7   PE6      VMK                   GPIO_VMK(outin)
   GPIO8   PB7      VMIO                  GPIO_VMIO(outin)

   GPIO9   PE8      ASY1 irq              GPIO_ASY_IRQ(0,outin)
   GPIO10  PA9      USB RC filter         
   GPIO11  PE10     ASY2 irq              GPIO_ASY_IRQ(1,outin)
   GPIO12  PE11     ASY3 irq              GPIO_ASY_IRQ(2,outin)

   GPIO13  PE12     drv_asy               GPIO_ASY(outin)
   GPIO14  PC13     drv_led               GPIO_LED(outin)
   GPIO15  PE14     drv_rcu               GPIO_RCU(outin)
   GPIO16  PD15     drv_kbd               GPIO_KBD(outin)

   GPIO17  PD3      Lost of event         GPIO_VMK_LOST(outin)
   GPIO18  PA6      Lost of vent          GPIO_VMIO_LOST(outin)
   GPIO19  PB8      1 KHz output signal   
   GPIO20  PC8      MMI task              GPIO_MMI(outin)

   GPIO21  PA10     producer task         GPIO_PROD(outin)
   GPIO22  PD10     consumer task         GPIO_CONS(outin)
   GPIO23  PC14     LIGHTS automaton      GPIO_LIGHTS(outin)
   GPIO24  PC15     SEND automaton        GPIO_SEND(outin)
*/

#define GPIO_BOOT(a)                   /* Not used on stm32wl.               */
#define GPIO_TIM(a)                    /* Not used on stm32wl.               */

#define GPIO_CPU(a)                    /* Not used on stm32wl.               */
#define GPIO_VMIT(a)                   /* Not used on stm32wl.               */
#define GPIO_VMK(a)                    /* Not used on stm32wl.               */
#define GPIO_VMIO(a)                   /* Not used on stm32wl.               */

#define IASY0(a)                       /* Not used on stm32wl.               */
#define IASY1(a)                       /* Not used on stm32wl.               */
#define IASY2(a)                       /* Not used on stm32wl.               */
#define GPIO_ASY_IRQ(n,a)              /* Not used on stm32wl.               */

#define GPIO_ASY(a)                    /* Not used on stm32wl.               */
#define GPIO_LED(a)                    /* Not used on stm32wl.               */
#define GPIO_RCU(a)                    /* Not used on stm32wl.               */
#define GPIO_KBD(a)                    /* Not used on stm32wl.               */

#define GPIO_VMK_LOST(a)               /* Not used on stm32wl.               */
#define GPIO_VMIO_LOST(a)              /* Not used on stm32wl.               */
#define GPIO_MMI(a)                    /* Not used on stm32wl.               */

#define GPIO_PROD(a)                   /* Not used on stm32wl.               */
#define GPIO_CONS(a)                   /* Not used on stm32wl.               */
#define GPIO_LIGHTS(a)                 /* Not used on stm32wl.               */
#define GPIO_SEND(a)                   /* Not used on stm32wl.               */

#endif                                 /* End of conditionnal compilation.   */
