/*****************************************************************************/
/*                                                                           */
/*  Hyperpanel OS                       File "include/armv7m.h"              */
/*  Constants and definitions for ARMV7 processors.                          */
/*                                                                           */
/*  Author : JYA                        Last modification : 07/15/2023       */
/*                                                                           */
/*****************************************************************************/


/*  Some explanations ========================================================


         1  Defines
         2  Structures and typedefs
         3  Public variables
         4  Prototypes of public procedures


*/


#ifndef ARMV7M_H                       /* If this file is not already        */
#define ARMV7M_H                       /* included, include it.              */


/*
    1    Defines
    ____________


         Les definitions presentes dans ce fichiers sont relatives aux
processeurs de la famille ARMv7-M. Il y a fort a parier que tout ou partie
des definitions s'applique aussi a d'autres processeurs proches.

*/

/*
    2    Structures and typedefs
    ____________________________
*/


typedef struct                         /* System control space (SCS)         */
  BEGIN                                /* ---------------------------------- */
    UINT            master           ; /* 000 Inutilise                      */
    UINT            ict              ; /* 004 Interrupt controller type      */
    UINT            actl             ; /* 008 Auxiliary control              */
    UINT            res[1]           ; /*                                    */
    UINT            systick[60]      ; /* 010 SysTick                        */
    UINT            nvic[768]        ; /* 100 Nested vector interrupt cont.  */
    UINT            cpuid            ; /* D00 CPU ID                         */
    UINT            ics              ; /* D04 Interrupt control state        */
    UINT            vto              ; /* D08 Vector table offset            */
    UINT            airc             ; /* D0C App. Interrupt/Reset Control   */
    UINT            sc               ; /* D10 System control                 */
    UINT            cc               ; /* D14 Configuration Control          */
    UINT            shp[3]           ; /* D18 System handler priority        */
    UINT            shcs             ; /* D24 System handler control & state */
    UINT            cfs              ; /* D28 Configurable fault status      */
    UINT            hfs              ; /* D2C Hard-fault status              */
    UINT            dfs              ; /* D30 Debug status                   */
    UINT            mma              ; /* D34 Memory manage address          */
    UINT            bfa              ; /* D38 Bus fault address              */
    UINT            afs              ; /* D3C Auxiliary fault status         */
    UINT            res2[16]         ; /*     Inutilise (pour CPUID)         */
    UINT            res3[2]          ; /*     Inutilise                      */
    UINT            cpac             ; /* D88 Coprocessor Access Control     */
    UINT            res4[1]          ; /*     Inutilise                      */
    UINT            mpu[24]          ; /* D90 Memory Protection Interrupt    */
    UINT            debug[68]        ; /* DF0 Debug                          */
    UINT            sti              ; /* F00 Software interrupt             */
  END               Scs_t            ; /* ---------------------------------- */

typedef struct                         /* Interrupt controller (NVIC)        */
  BEGIN                                /* -----------------------------------*/
    UINT           ise    [16]       ; /* 000 Set-Enable    registers        */
    UINT           pad01  [16]       ; /* 040 Unused                         */
    UINT           ice    [16]       ; /* 080 Clear-Enable  registers        */
    UINT           pad02  [16]       ; /* 0C0 Unused                         */
    UINT           isp    [16]       ; /* 100 Set-Pending   registers        */
    UINT           pad03  [16]       ; /* 140 Unused                         */
    UINT           icp    [16]       ; /* 180 Clear-Pending registers        */
    UINT           pad04  [16]       ; /* 1C0 Unused                         */
    UINT           iab    [16]       ; /* 200 Active-Bit    registers        */
    UINT           pad05  [48]       ; /* 240 Unused                         */
    UINT           ip    [124]       ; /* 300 Priority      registers        */
  END              Nvic_t            ; /* -----------------------------------*/

typedef struct                         /* NVIC Software IRQ reference entry  */
  BEGIN                                /* -----------------------------------*/
    CHAR           *name             ; /* Reference name                     */
    UCHAR          num               ; /* Software number                    */
    UCHAR          allocsys          ; /* 0: User allocated 1:Sys allocated  */
  END              Ref_nvic          ; /* -----------------------------------*/

/*
    3    Public variables
    _____________________
*/

extern volatile Nvic_t *nvic         ; /* NVIC registers base address        */

/*
    4    Prototypes of public procedures
    _____________________________________
*/

IWORD softirq_nvic    (IWORD,IWORD                                          );

#endif                                 /* End of ARMV7M_H.                   */
