/*****************************************************************************/
/*                                                                           */
/*  Hyperpanel OS                       File "include/vmk_tim.h"             */
/*  Constants and definitions for IT driver.                                 */
/*                                                                           */
/*  Author : JYA                        Last modification : 04/04/2021       */
/*                                                                           */
/*****************************************************************************/


/*  Some explanations ========================================================


         1  Defines

         2  Structures and typedefs

         3  Public variables
            3.1  64 bits counter of milliseconds elapsed since boot
            3.2  Seconds and milliseconds since January  1st, 2000
            3.3  Seconds and milliseconds since boot

         4  Prototypes of public procedures
            4.1  Interface procedures
            4.2  Low-level procedures interface

*/


#ifndef DRV_TIM_H                      /* If this file is not already        */
#define DRV_TIM_H                      /* included, include it.              */


/*
    1    Defines
    ____________
*/


/*
    2    Structures and typedefs
    ____________________________
*/


/*
    3    Public variables
    _____________________


    3.1  64 bits counter of milliseconds elapsed since boot
    -------------------------------------------------------
*/

extern volatile UINT  ticks_low      ; /* Count of elapsed msec since boot   */
extern volatile UINT  ticks_high     ; /* 64 bits extension of "temps"       */


/*  3.2  Seconds and milliseconds since January  1st, 2000
    ------------------------------------------------------
*/

extern volatile UINT  sec            ; /* Count of seconds since 01/01/2000  */
extern volatile UINT  msec           ; /* Count of milliseconds 0..999       */


/*  3.3  Seconds and milliseconds since boot
    ----------------------------------------
*/

extern volatile UINT  bootsec        ; /* Count of seconds since boot        */
extern volatile UINT  bootmsec       ; /* Count of milliseconds 0..999       */


/*
    4    Prototypes of public procedures
    ____________________________________


    4.1  Interface procedures
    -------------------------
*/

void  tim_get             (INT*,INT*,INT*,INT*,INT*,INT*,INT*                );
void  tim_set             (INT,INT,INT,INT,INT,INT,INT,INT*                  );
void  tim_get_ticks       (UINT*,UINT*                                       );
void  tim_get_usec        (UINT*                                             );


/*  4.2  Low-level procedures interface
    -----------------------------------
*/

IWORD timget_sec_msec     (UINT*,UINT*                                       );
IWORD timset_sec_msec     (UINT, UINT                                        );
IWORD tim_fastproc        (UINT, IWORD (*)()                                 );
void  timget_hdw_timer    (UINT*,UINT*                                       );
void  tim_uwait           (UINT                                              );


#endif                                 /* End of DRV_TIM_H                   */
