/*****************************************************************************/
/*                                                                           */
/*  Hyperpanel OS                       File "include/tsk_fsmlib.h"          */
/*  Interface definitions for "tsk_fsmlib.h"                                 */
/*                                                                           */
/*  Author : JYA                        Last modification : 10/18/2022       */
/*                                                                           */
/*  (c) Copyright Hyperpanel Lab, All Rights Reserved                        */
/*  IDDN.FR.001.320001.*.S.P.1990.000.10000                                  */
/*                                                                           */
/*****************************************************************************/

/*  Some explanations ========================================================


         1  Defines
         2  Structures and typedefs
            2.1  One entry of a table for event conversion
            2.2 One entry of a transition table
         3  Public variables
         4  Prototypes of public procedures
         5  RAM data to be included in space_appcont_xyz.c
*/

#ifndef TSKFSMLIB_B                    /* If this file is not already        */
#define TSKFSMLIB_B                    /* included, include it.              */


/*
    1    Defines
    _____________
*/

#define  FSM_ERROR            -1       /* Treatment has set an error code    */


/*
    2    Structures and typedefs
    ____________________________


    2.1  One entry of a table for event conversion
    ----------------------------------------------
*/

typedef struct                         /* Run treatment of incoming event    */
  BEGIN                                /* ---------------------------------- */
    USHORT         codeevt           ; /* Module name address                */
    INT            *iod              ; /* Address of the "iod" value         */
    IWORD          (*conv)()         ; /* Procedure to be called             */
    INT            param             ; /* Parameter for "proc"               */
  END              Conv_evt          ; /* ---------------------------------- */

/*  2.2 One entry of a transition table
    -----------------------------------
*/

typedef struct                         /* One menu transition table line     */
  BEGIN                                /* ---------------------------------- */
    SHORT          evtcode           ; /* Awaited code event (opt value)     */
    int  (*trt1)() ; UINT  p1        ; /* 1st treatment and its parameter    */
    int  (*trt2)() ; UINT  p2        ; /* 2nd treatment and its parameter    */
    int  (*trt3)() ; UINT  p3        ; /* 3rd treatment and its parameter    */
    int  (*trt4)() ; UINT  p4        ; /* 4th treatment and its parameter    */
    int  (*trt5)() ; UINT  p5        ; /* 5th treatment and its parameter    */
  END              M_trans           ; /* ---------------------------------- */

/*
    3    Public variables
    _____________________
*/

extern INT fsm_state                 ; /* FSM state variable                 */
extern INT fsm_evt                   ; /* FSM event code                     */
extern INT fsm_cond                  ; /* FSM condition code value           */
extern INT fsm_condret               ; /* FSM return condition code value    */
extern INT fsm_err                   ; /* FSM global error code              */


/*
    4    Prototypes of public procedures
    ____________________________________
*/

IWORD fsm_conv_evt        (S_evt*, const Conv_evt*                           );
void  fsm_run             (const M_trans *const[]                            );

IWORD empt                (int                                               );
IWORD next                (int                                               );
IWORD sret                (int                                               );
IWORD scon                (int                                               );

#endif                                 /* End of VMK_IOLIB_H                 */


