/*****************************************************************************/
/*                                                                           */
/*  Hyperpanel OS -                     File "os/vmk/lib/vmk_iolib.c"        */
/*  ...                                                                      */
/*                                                                           */
/*  Author : JYA                        Last modification : 10/18/2022       */
/*                                                                           */
/*****************************************************************************/


/*  Some explanations ========================================================


    1    Purpose of this module
    ___________________________


         This module provides to applicative software a simplified interface
to "drv.c", which is limited to use of devices. We also have a limited number
of provided functions and so the size of the documentation to be read is
reduced, compared to "drv.c" user manual document.

         Bla Bla
   
    Some rules for on-line documentation generation :

    #PRIVATE   ... #END     Not generated in on-line documentation.
    #DIAGRAM   ... #END     Diagram, drawing, schematic.
    #CODE      ... #END     Source code example (ex: typedef).
    #ARRAY     ... #END     Array (ex: explanation following typedef).
    #INDEX     ... #END     Sub-chapters list, procedure list.
    #PCODE     ... #END     Pseudo-code.
    #PROCDESCR ... #END     Procedure interface description.

#PRIVATE ----------------------------------------------------------------------

NOTES ========================================================================

CODE =========================================================================

DOC ==========================================================================

TESTS ========================================================================

EVOLUTIONS ===================================================================


#END --------------------------------------------------------------------------



    2    User manual
    ________________


         2.1  What is a driver?
              2.1.1  Driver and devices
              2.1.2  Writing to a device
              2.1.3  Reading from a device
              2.1.4  Device descriptors
              2.1.5  Blocking and non-blocking write and reads

         2.2  The "vmkiolib.c" library
              2.2.1  The "drv.c" API and the "vmk_iolib.c" library
              2.2.2  Less functions
              2.3.3  Less function parameters
              2.3.4  Drivers and MODD identifiers
              2.3.5  Devices and channel IOD identifiers
              2.3.6  Sub-channels and sub-channel IOD identifiers
              2.3.7  Mixing "drv.c" and "vmk_iolib.c" calls
              2.3.8  "vmk_iolib.c" calls to "drv.c" AIP functions
               
         2.3  How to use the "vmk_iolib.c" library
              2.3.1  Getting the driver names
              2.3.2  Getting the device names for a driver
              2.3.3  Opening and closing a device
              2.3.4  Write and send to a device
              2.3.5  Reading from a device
              2.3.6  Setting and getting control parameters

         2.4  Get information procedures
              2.4.1  List of procedures
              2.4.2  io_list_drivers
              2.4.3  io_list_devices
              2.4.5  io_list_tags
              2.4.6  io_tag_number
              2.4.7  io_tag_name

         2.5  Initialization procedures
              2.5.1  List of procedures
              2.5.2  io_open
              2.5.3  io_close
              2.5.4  io_alloc_sub
              2.5.5  io_free_sub

         2.6  Write and read procedures
              2.6.1  List of procedures
              2.6.2  io_write
              2.6.3  io_send
              2.6.4  io_read
              2.6.5  io_receive

         2.7  Set and read control parameter procedures
              2.7.1  List of procedures
              2.7.2  io_setval
              2.7.3  io_getval

         2.8  Misceleanous procedures
              2.8.1  List of procedures
              2.8.2  io_bufsize

    2.1  xxxxxxx
         Bla bla

    3    Porting guide
    __________________


         3.1  xxxxxxx
              3.1.1  xxxxxx

    3.1  xxxxxxx
         Bla bla


    4    Include files and external references
    __________________________________________
*/


#include <stdtyp.l>                    /* Hyperpanel definitions & constants */
#include <event.l>                     /* IND_REPORT                         */
#include <aut.h>                       /* S_evt                              */
#include <vmio.h>                      /* BAD_PARAMETER                      */
#include <vmk.h>                       /* "vmk.c" prototypes and constants   */
#include <drv.h>                       /* "drv.c" prototypes and constants   */
#include <tsk_iolib.h>                 /* Mod_def                            */
#include <text.h>                      /* hsprintf                           */
                                       /*                                    */
#include <systemm.txt>                 /* WAIT_CODERES                       */
#include <messagem.txt>                /* BAD_NOMP                           */

#ifdef TEST                            /* For unitary tests.                 */
#endif                                 /* End of ifdef TEST.                 */



/* External reference from space_appcont_xyz.c" .............................*/

extern const Mod_def iolib_moddef[] ; /* Table of requested modules          */
                                      /*                                     */
extern INT           iolib_modd []  ; /* Storage for the MODD's              */
extern UINT          iolib_rteid[]  ; /* Storage for route identifiers       */
extern UCHAR         iolib_iodnb[]  ; /* Count of allocated devices          */
extern INT           iolib_iod  []  ; /* Storage for the IOD's               */
                                      /*                                     */
extern const INT     iolib_nbmodd   ; /* Size of iolib_moddef[] iolib_modd[] */
extern const INT     iolib_nbiod    ; /* Size of "iolib_iod[]"               */


extern const Sys_driver * const vmos_driver[]; /* Table des desc. drivers    */
extern const USHORT  vmos_nbdriver   ; /* Nombre d'elements de vmos_driver   */

/* External reference from ptk_lib.c ........................................*/

extern IWORD         vlgv            ; /* Maximum string size in bytes       */

/*
    5    Inner workings of the module
    _________________________________


         5.1  xxxxx
              5.1.1  xxxxx

         5.2  xxxxx
              5.2.1  xxxx


    6    Exported (public) variables
    ________________________________
*/

/*
    7    Internal constants of the module
    _____________________________________
*/

/*
    8    Internal data types of the module
    ______________________________________
*/

static IWORD io_chk_driver   (const CHAR*,INT*,INT*                          );
static IWORD io_get_iodidx   (INT*                                           );
static IWORD io_chk_iod      (INT,INT,INT*,INT*                              );

/*
    9    Internal (private) variables of the module
    _______________________________________________
*/

/*
    10   Beginning of the code
    __________________________

io_list_driver     Get the list of driver names
io_list_devices    Get the list of device names for one driver
io_list_tags       Get the list of tag names for one driver
io_tag_number      Get the number of a given tag name
io_tag_name        Get the name of a given tag number

io_open            Intializes the driver, allocate and opens a device
io_close           Close and free one device, terminate the driver
io_alloc_sub       Allocate one sub-channel of one opened device
io_free_sub        Free one sub-channel

io_write           Blocking write of of data buffer
io_send            No blocking write of one data buffer
io_read            Blocking read
io_receive         Gives several read tokend

io_setval          Set/writes a taglist
io_getval          Gets/reads a tag list

io_bufsize         Return the "len" needed buffer size foir "lg" data bytes

io_get_iodidx      Subroutine which allocate an "iolib_iod[]" free entry
io_chk_driver      Check that a driver name is known to "vmk_iolib"
io_chk_iod         Check that an iod has been allocated by "vmk_iolib"

*/


/*  Procedure io_list_drivers -----------------------------------------------*/
/*
    Purpose : Get the list of all the available character drivers

    Author : JYA                                 Last modification : 10/23/2022

...............................................................................

    Calling example : io_list_drivers( list, size, nb )

    Input     : list     char*      Buffer for the names
                size     int        Size in bytes of "list"

    Output    : nb       int*       Count of returned names
                valproc  IWORD      Error code

    Notes : 

...............................................................................
*/

IWORD  PROCEDURE io_list_drivers(char *list, int size, int *nb)

  BEGIN                                /* Begin of this procedure            */
 
/*  Declaration of local variables of this procedure ........................*/

    IWORD                 r          ; /* Return code of the procedure       */
    INT                   len        ; /* Count of copied bytes              */
    IWORD                 i          ; /* Index in "iolib_moddef[]"          */
    const Mod_def         *pmd       ; /* Address of "iolib_moddef[i]"       */
    register const CHAR   *s         ; /* Copy source address                */
    register CHAR         *d         ; /* Copy destination addres            */
    INT                   moidx, maj ; /* "io_chk_driver" returned values    */

/*  Beginning of the code ...................................................*/

#ifdef TEST                            /* Begin conditional compilation      */
    procdeb("io_list_drivers")       ; /* Message                            */
#endif                                 /* End   conditional compilation      */

/*****************************************************************************
 * Step  1 : Initialization of local variables:                              *
 * -------   - We will use the "d" local variable as the character write     *
 *             pointer to the "list" output string/buffer. We initialize     *
 *             "d" to "list".                                                *
 *           - We clear the "len" count of written bytes to "list"           *
 *           - We clear the "*nb" count of written driver names              *
 *           - In the very unlikely event we had 0 drivers, we put a 0 C     *
 *             string terminator in "list[0]"                                *
 *           --------------------------------------------------------------- *
 *           We could directly use "list" as the write pointer. However,     *
 *           function parameters are not always registers, depending on the  *
 *           C compiler.                                                     *
 *****************************************************************************/

    d    = list                      ; /* Set the "d" write pointer to "list"*/
    len  = 0                         ; /* Clear count of written bytes       */
    *nb  = 0                         ; /* Init count of written names        */
    *d   = 0                         ; /* In case of empty string            */


/*****************************************************************************
 * Step  2 : We loop on the "iolib_nbmodd" entries of the "iolib_moddef[]"   *
 * -------   table and we copy the "modname" fields to "list". We add a      *
 *           line-feed at the end of each copied name. We also use the loop  *
 *           to check that each "modname" can also be found in the           *
 *           "vmos_driver[]" table, which should be always the case. For     *
 *           each entry of "iolib_moddef[]" we :                             *
 *           - Check that the "pmd->name" string can be found in the         *
 *             "vmos_driver[]" table. We call "io_chk_driver" to do that. If *
 *             the "pmd->name" is not found, we have a big problem, data     *
 *             in "space_appcont_xyz.c" is incorrect. In case or error, the  *
 *             "io_chk_driver" subroutine returns -1.                        *
 *           - Increment the "len" count of wrtiien bytes with "Strlen(s)+1" *
 *             (Do not forget the trailing '\n"  separator byte).            *
 *           - If "len" is greater than "size" we cannot copy the new name   *
 *             into "list", so we exit with STRING_TOO_SHORT.                *
 *****************************************************************************/

    for( i = 0, pmd = iolib_moddef;    /* We loop with "iolib_nbmodd" entries*/
         i LT iolib_nbmodd        ;    /* of the "moddef[]" const table,     */
         i++, pmd++            )       /* using "i" as an index and "pmd"    */
      BEGIN                            /* as a pointer to "moddef[i]"        */
        s = pmd->modname             ; /* The address of the driver name     */
                                       /*                                    */
        r = io_chk_driver(s,           /* Check that the "s" driver name     */
                       &moidx,&maj)  ; /* can be found in the "vmos_driver"  */
        if ( r )                       /* table. If this is not the case,    */
          goto exit                  ; /* exit with the -1 error code        */
                                       /*                                    */
        len += Strlen(s) + 1         ; /* Increment the count of written     */
                                       /* bytes, including the '\n' byte     */
        if ( len GT size )             /* If we do not have enough room      */
          BEGIN                        /* any more, we do not copy the new   */
            r = STRING_TOO_SHORT     ; /* driver name, we exit with the      */
            goto exit                ; /* STRING_TOO_SHORT error code        */
          END_IF                       /*                                    */
                                       /*                                    */
        while (*s) *d++ = *s++       ; /* Copy the name, without the 0 byte  */
                                       /* terminator.                        */
        *d++ = '\n'                  ; /* Add a '\n' separator               */
                                       /*                                    */
        (*nb)++                      ; /* Increment count of written names   */
      END_FOR                          /*                                    */


/*****************************************************************************
 * Step  3 : We successfully copied the "iolib_nbmodd" names. The last '\n"  *
 * -------   is overwritten with a 0 C string terminator, at address "d-1".  *
 *           In the very unlikely event we get an empty string (no driver !!)*
 *           we have "d = list" and we do not write anything at "d-1"        *
 *****************************************************************************/

    exit                             : /* Procedure exit label               */

    if ( len )                         /* If the string is not an empty one  */
     *(d-1) = 0                      ; /* overwrite the last '\n' with a 0   */

#ifdef TEST                            /* Begin conditional compilation      */
    procfin("io_list_drivers")       ; /* Message                            */
#endif                                 /* End   conditional compilation      */

    return( r )                      ; /* Return code of the procedure       */

  END_PROC                             /* End of the procedure               */


/*  Procedure io_list_devices -----------------------------------------------*/
/*
    Purpose : Get the list of devices names for a given driver

    Author : JYA                                 Last modification : 10/23/2022

...............................................................................

    Calling example : io_list_devices( driver, list, size, nb )

    Input     : driver   char*      Involved driver name
                list     char*      Buffer for the names
                size     int        Size in bytes of "list"

    Output    : nb       int*       Count of returned names
                valproc  IWORD      Error code

    Notes : 

...............................................................................
*/

IWORD  PROCEDURE io_list_devices(const char *driver, char *list, int size,
                                 int *nb                                  )

  BEGIN                                /* Begin of this procedure            */
 
/*  Declaration of local variables of this procedure ........................*/

    IWORD                 r          ; /* Return code of the procedure       */
    INT                   moidx, maj ; /* "io_chk_driver" returned values    */

/*  Beginning of the code ...................................................*/

#ifdef TEST                            /* Begin conditional compilation      */
    procdeb("io_list_devices")       ; /* Message                            */
#endif                                 /* End   conditional compilation      */

/*****************************************************************************
 * Step  1 : Check that the "driver" name is known by the present module,    *
 * -------   that "driver" can be found in the "iolib_moddef[]" const table. *
 *           We call "io_chk_driver", which also checks that this name can   *
 *           also be found in the "vmos_driver[]" table. Return code is :    *
 *           - 0  if "driver" has been found in both "mod_ref[]" and         *
 *             "vmos_driver[]" tables.                                       *
 *           - BAD_NOMP is "driver has not been found in "iolib_moddef[]"    *
 *           - -1 if "driver" has been found in "iolib_moddef[]" but not in  *
 *             the "vmos_driver[]" table. In such a case, configuration data *
 *             of "space_appcont_xyz.c" is wrong.                            *
 *****************************************************************************/

    *list = 0                        ; /* In case of error, set a 0 C string */
    *nb   = 0                        ; /* terminator and clear the count     */
                                       /*                                    */
    r = io_chk_driver ( driver ,       /* Check that "driver" string can be  */
                     &moidx, &maj)   ; /* found in both "iolib_moddef[]" and */
    if ( r )                           /* "vmos_driver". If this is not the  */
      goto exit                      ; /* case we exit with BAD_NOMP or -1   */

/*****************************************************************************
 * Step  2 : Call the "drv.c" API procedure, in our case "drv_list_names".   *
 * -------   This function is intended to be used with character or block    *
 *           drivers. We do not ask for the =FREE/=ALLOC/=OPEN string that   *
 *           may follow each device name.                                    *
 *****************************************************************************/

    vlgv = size                      ; /* The output string size in bytes    */
                                       /*                                    */
    r = drv_list_names( (CHAR*)driver, /* module   driver name (string)      */
                        list  ,        /* list     Output string buffer      */
                        nb    ,        /* nb       Count of returned names   */
                        0       )    ; /* state  0 Do not return states      */

    exit                             : /* Exit label                         */

#ifdef TEST                            /* Begin conditional compilation      */
    procfin("io_list_devices")       ; /* Message                            */
#endif                                 /* End   conditional compilation      */

    return( r )                      ; /* Return code of the procedure       */

  END_PROC                             /* End of the procedure               */


/*  Procedure io_list_tags --------------------------------------------------*/
/*
    Purpose : Get the list of tags for one driver

    Author : JYA                                 Last modification : 10/23/2022

...............................................................................

    Calling example : 

    Input     : x        IWORD      C

    Output    : 

    Notes : 

...............................................................................
*/

IWORD  PROCEDURE io_list_tags(const char *driver, char *list, int size, 
                              int *nb                                  )

  BEGIN                                /* Begin of this procedure            */
 
/*  Declaration of local variables of this procedure ........................*/

    IWORD          r                 ; /* Return code of the procedure       */
    INT            moidx, maj        ; /* "io_chk_driver" returned values    */

/*  Beginning of the code ...................................................*/

#ifdef TEST                            /* Begin conditional compilation      */
    procdeb("io_list_tags")          ; /* Message                            */
#endif                                 /* End   conditional compilation      */

/*****************************************************************************
 * Step  1 : Check that the "driver" name is known by the present module,    *
 * -------   that "driver" can be found in the "iolib_moddef[]" const table. *
 *           We call "io_chk_driver", which also checks that this name can   *
 *           also be found in the "vmos_driver[]" able. Return code is :     *
 *           - 0  if "driver" has been found in both "mod_ref[]" and         *
 *             "vmos_driver[]" tables.                                       *
 *           - BAD_NOMP is "driver has not been found in "iolib_moddef[]"    *
 *           - -1 if "driver" has been found in "iolib_moddef[]" but not in  *
 *             the "vmos_driver[]" table. In such a case, configuration data *
 *             in "space_appcont_xyz.c" is wrong.                            *
 *****************************************************************************/

    *list = 0                        ; /* In case of error, set a 0 C string */
    *nb   = 0                        ; /* terminator and clear the count     */
                                       /*                                    */
    r = io_chk_driver ( driver ,       /* Check that "driver" string can be  */
                     &moidx, &maj)   ; /* found in both "iolib_moddef[]" and */
    if ( r )                           /* "vmos_driver". If this is not the  */
      goto exit                      ; /* case we exit with BAD_NOMP or -1   */

/*****************************************************************************
 * Step  2 : Get the list of tags, using the "drv_list_tag" API function     *
 * -------   of "drv.c". It has to be noted that the buffer size is given    *
 *           by the "vlgv" global variable, that we must set to "size"       *
 *           before calling "drv_list_tag"                                   *
 *****************************************************************************/

    vlgv = size                      ;

    r = drv_list_tags((CHAR*)driver ,
                       list   ,
                       nb      )     ;

    exit                             : /* Exit label                         */

#ifdef TEST                            /* Begin conditional compilation      */
    procfin("io_list_tags")          ; /* Message                            */
#endif                                 /* End   conditional compilation      */

    return( r )                      ; /* Return code of the procedure       */

  END_PROC                             /* End of the procedure               */


/*  Procedure io_tag_number -------------------------------------------------*/
/*
    Purpose : Get the tag number of a given tag name.

    Author : JYA                                 Last modification : 10/23/2022

...............................................................................

    Calling example : io_tag_number ( driver, tagname, tagnum )

    Input     : driver   char*      Driver name
                tagname  char*      tag name

    Output    : tagnum   int*       Tag number of "tagname"
                valproc  IWORD      Procedure return code

    Notes : 

...............................................................................
*/

IWORD  PROCEDURE io_tag_number(const char *driver, const char *tagname,
                               int *tagnum                             )

  BEGIN                                /* Begin of this procedure            */
 
/*  Declaration of local variables of this procedure ........................*/

    IWORD          r                 ; /* Return code of the procedure       */
    INT            moidx, maj        ; /* "io_chk_driver" returned values    */
    INT            sub               ; /* 0:channel 1:sub-channel            */

/*  Beginning of the code ...................................................*/

#ifdef TEST                            /* Begin conditional compilation      */
    procdeb("io_tag_number")         ; /* Message                            */
#endif                                 /* End   conditional compilation      */

/*****************************************************************************
 * Step  1 : Check that the "driver" name is known by the present module,    *
 * -------   that "driver" can be found in the "iolib_moddef[]" const table. *
 *           we call "io_chk_driver", which also checks that this name can   *
 *           also be found in the "vmos_driver[]" table. Return code is :    *
 *           - 0  if "driver" has been found in both "mod_ref[]" and         *
 *             "vmos_driver[]" tables.                                       *
 *           - BAD_NOMP is "driver has not been found in "iolib_moddef[]"    *
 *           - -1 if "driver" has been found in "iolib_moddef[]" but not in  *
 *             the "vmos_driver[]" table. In such a case, configuration data *
 *             in "space_appcont_xyz.c" is wrong.                            *
 *****************************************************************************/

    *tagnum = -1                     ; /* Let'a assume we will fail          */
                                       /*                                    */
    r = io_chk_driver ( driver ,       /* Check that "driver" string can be  */
                     &moidx, &maj)   ; /* found in both "iolib_moddef[]" and */
    if ( r )                           /* "vmos_driver". If this is not the  */
      goto exit                      ; /* case we exit with BAD_NOMP or -1   */


/*****************************************************************************
 * Step  2 : Convert the "tagname" string to the "tagnum" corresponding      *
 * -------   tag number, using the "drv_tag_number" API function from        *
 *           "drv.c"                                                         *
 *****************************************************************************/

    r = drv_tag_number((CHAR*)driver,  /* module  Our driver name            */
                       (CHAR*)tagname, /* tag     Name of the tag            */
                       tagnum        , /* num     Tag order number           */
                       &sub         ); /* sub     0:channel 1:sub-channel    */

    exit                             : /* Exit label                         */

#ifdef TEST                            /* Begin conditional compilation      */
    procfin("io_tag_number")         ; /* Message                            */
#endif                                 /* End   conditional compilation      */

    return( r )                      ; /* Return code of the procedure       */

  END_PROC                             /* End of the procedure               */


/*  Procedure io_tag_name ---------------------------------------------------*/
/*
    Purpose : Get the tag name if a given tag number.

    Author : JYA                                 Last modification : 10/23/2022 

...............................................................................

    Calling example : io_tag_name( driver, tagnum, tagname, size )

    Input     : driver   char*      Involved driver name
                tagnum   int        Tag number
                tagname  char*      Tag name buffer
                size     int        "tagname" size in bytes

    Output    : valproc  IWORD      Procedure error code

    Notes : 

...............................................................................
*/

IWORD  PROCEDURE io_tag_name(const char* driver, int tagnum,
                             char *tagname, int size        )

  BEGIN                                /* Begin of this procedure            */
 
/*  Declaration of local variables of this procedure ........................*/

    IWORD          r                 ; /* Return code of the procedure       */
    INT            moidx, maj        ; /* "io_chk_driver" returned values    */

/*  Beginning of the code ...................................................*/

#ifdef TEST                            /* Begin conditional compilation      */
    procdeb("io_tag_name")           ; /* Message                            */
#endif                                 /* End   conditional compilation      */

/*****************************************************************************
 * Step  1 : Check that the "driver" name is known by the present module,    *
 * -------   that "driver" can be found in the "iolib_moddef[]" const table. *
 *           We call "io_chk_driver", which also checks that this name can   *
 *           also be found in the "vmos_driver[]" table. Return code is :    *
 *           - 0  if "driver" has been found in both "mod_ref[]" and         *
 *             "vmos_driver[]" tables.                                       *
 *           - BAD_NOMP is "driver has not been found in "iolib_moddef[]"    *
 *           - -1 if "driver" has been found in "iolib_moddef[]" but not in  *
 *             "vmos_driver[]" table. In such a case, "space_appcont_xyz.c"  *
 *             is wrong.                                                     *
 *****************************************************************************/

    *tagname = 0                     ; /* Let'a assume we will fail          */
                                       /*                                    */
    r = io_chk_driver ( driver ,       /* Check that "driver" string can be  */
                     &moidx, &maj)   ; /* found in both "iolib_moddef[]" and */
    if ( r )                           /* "vmos_driver". If this is not the  */
      goto exit                      ; /* case we exit with BAD_NOMP or -1   */


/*****************************************************************************
 * Step  2 : Convert the "tagnum" tag order number to the "tagname"          *
 * -------   corresponding string, using the "drv_tag_name" API function     *
 *           from "drv.c"                                                    *
 *****************************************************************************/

    vlgv = size                      ; /* Give the string size in bytes      */
                                       /*                                    */
    r = drv_tag_name((CHAR*)driver ,   /* module  Name of the driver         */
                     tagnum        ,   /* num     Tag order number           */
                     tagname        ); /* tag     Name of the tag            */

    exit                             : /* Exit label                         */

#ifdef TEST                            /* Begin conditional compilation      */
    procfin("io_tag_name")           ; /* Message                            */
#endif                                 /* End   conditional compilation      */

    return( r )                      ; /* Return code of the procedure       */

  END_PROC                             /* End of the procedure               */


/*  Procedure io_open -------------------------------------------------------*/
/*
    Purpose : Initializes the driver if needed, create a route for IND_REPORT
              if needed, allocate one device and finally open the device.

    Author : JYA                                 Last modification : 10/18/2022

...............................................................................

    Calling example : io_open( modname, devname, taglist, iod )

    Input     : modname  CHAR*      Driver module name
                devname  CHAR*      Device name
                taglist  CHAR*      List of tags for "drv_open_device"

    Output    : iod      INT*       IO channel identifier

    Notes : 

...............................................................................
*/

IWORD  PROCEDURE io_open(const char *modname, const char *devname, 
                         const char *taglist,int *iod            )

  BEGIN                                /* Begin of this procedure            */
 
/*  Declaration of local variables of this procedure ........................*/

    IWORD          r                 ; /* Return code of the procedure       */
    IWORD          moidx             ; /* Loop counter/iolib_modd[] index    */
    IWORD          ioidx             ; /* "iolib_iod[]" index                */
    const Mod_def  *pmd              ; /* Pointer in "iolib_moddef[]" table  */
    INT            major             ; /* Driver major number << 16          */
    INT            maxres , minres   ; /* Values for add_uroute              */
    INT            maxcode, mincode  ; /* Values for add_uroute              */
    CHAR           name[32]          ; /* For "drv_alloc_device"             */

/*  Beginning of the code ...................................................*/

#ifdef TEST                            /* Begin conditional compilation      */
    procdeb("io_open")               ; /* Message                            */
#endif                                 /* End   conditional compilation      */

/*****************************************************************************
 * Step  1 : Check that we have at least one available entry in the          *
 * -------   "iolib_iod[]" IOD storage table. We call the "io_get_iodidx"    *
 *           subroutine which returns the first "ioidx" free entry index.    *
 *           If the table is full, function return value is NOT_ENOUGH_MEM   *
 *****************************************************************************/

    r = io_get_iodidx ( &ioidx )     ; /* Get the first free entry index     */
                                       /* in "iolib_iod[]"                   */
    if ( r )                           /* if we get NOT_ENOUGH_MEM, we exit  */
      goto exit                      ; /* with that error code value         */


/*****************************************************************************
 * Step  2 : Search for the "modname" name in the "modlist" const table      *
 * -------   and exit with BAD_NOMP if this name is not found in the const   *
 *           "iolib_moddef[]" table. "moidx" is an index in "iolib_moddef[]" *
 *           and ""pmd" is a pointer to "iolib_moddef[moidx]"                *
 *           --------------------------------------------------------------- *
 *           We remind that iolib_moddef" table is an array of "Mod_def".    *
 *           This const array should be defined in "space_appcont_xyz.c".    *
 *           We also remind that "space_appcont.xyz.c" includes "vm_iolib.c" *
 *           in wich a "const INT" variable named "nbmoddef" is defined,     *
 *           its value beeing the "iolib_moddef[]" table size.               *
 *****************************************************************************/

    r = BAD_NOMP                     ; /* Let's assume we will fail          */
                                       /*                                    */
    for (moidx=0, pmd=iolib_moddef ;   /* We loop on the "nbmoddef" entries  */
         moidx LT iolib_nbmodd     ;   /* of the "iolib_moddef[]" table,     */
         moidx++, pmd++              ) /* "moidx" is the table index whilst  */
      BEGIN                            /* "pmd" is "&iolib_moddef[moidx]"    */
        if(Strcmp(modname,             /* We search for the "modname" input  */
                  pmd->modname) EQ 0)  /* parameter string, if we find it    */
          goto step3                 ; /* we are so happy, we jump to 2 !!   */
      END_FOR                          /* Not found, continue looping        */
                                       /*                                    */
    goto exit                        ; /* We are so sad, exit with BAD_NOMP  */
 

/*****************************************************************************
 * Step 3 : If the driver is not yet initialized, if the "iolib_modd[moidx]" *
 * ------   global variable value is now 0, then we have to send a REQ_INIT  *
 *          initialization request event to the VMIO driver FSM. To do this, *
 *          we call the "drv_init_driver" procedure. The VMIO driver FSM     *
 *          responds with a RESP_INIT event, we here unschedule until        *
 *          reception of this of this response event                         *
 *****************************************************************************/

    step3                            : /* Driver initialization if needed    */
                                       /*                                    */
    if( iolib_modd[moidx] )            /* If the module is already           */
      goto step5                     ; /* initialized, skip step3 and step4  */
                                       /*                                    */
    r = drv_init_driver(               /* Sends a REQ_INIT                   */
                  (CHAR*)modname     , /* Name of the driver module          */
                  0                  , /* Node number = local (0)            */
                  1                  , /* Reqid                              */
                  iolib_modd + moidx); /* Module identifier MODD             */
                                       /*                                    */
    if ( r ) goto exit               ; /* Exit if error                      */


/*****************************************************************************
 * Step 4 : If necessary, create one and one only user route per driver in   *
 * -------  order to receive the unsollicited events, that are in our case:  *
 *          - IND_REPORT for the ASY and CAN drivers                         *
 *          - IND_REPORT for the GPIO driver                                 *
 *          - DRV_KEY_PRESS and DRV_KEY_RELEASE for the KBD driver           *
 *          The ADC and the LED do not generate unsollicited events. We use  *
 *          the "iolib_moddef[]" table that contains "mincode" and "maxcode" *
 *          values to be given to "add_route". When no route has to be       *
 *          created, we put a -1 for the "mincode" value (LED and DAC)       *
 *          - With the KBD, we may have two event codes that are             *
 *            DRV_KEY_PRESS (0) and DRV_KEY_RELEASE (1). The smallest one is *
 *            used as "mincode" (0) and the greatest one is "maxcode" (1)    *
 *            The "reserve" field of those events is the keycode and we      *
 *            accept any value, therefore we use "minres = maxres = -1"      *
 *          - With ASY and GPIO, events are IND_REPORT and we use in that    *
 *            case "mincode = maxcode = IND_REPORT". The "reserve" field of  *
 *            the IND_REPORT is the device IOD (ASY) or the sub-channel IOD  *
 *            (GPIO). We extract the "major number" from the MODD and we     *
 *            compute "minres" = "major" + 0x01000000 and "maxres" = "major" *
 *            + 0x0100FFFF. The "major" is bits b23..16 in both MODD and IOD *
 *            So we have a route that will "match" with all the ASY or LED   *
 *            events, whatever are the minor number and the sub-channel      *
 *          ---------------------------------------------------------------- *
 *          The structure of any MODD is:                                    *
 *          - Bits b31..28 are 0000  for an MODD or IOD                      *
 *          - Bits b27..24 are 1111  for a MODD                              *
 *          - Bits b23..16 are the major number                              *
 *          - Bits b15..0  are the selector (MODD index)                     *
 *          ---------------------------------------------------------------- *
 *          With IND_REPORTS events, the "reserve" field is the IOD. The     *
 *          structure of any IOD is as follows:                              *
 *          - Bits b31..28 are 0000  for an MODD or IOD                      *
 *          - Bits b27..24 are 0001  for a character driver                  *
 *          - Bits b23..16 are the major number                              *
 *          - Bits b15..0  are the minor number                              *
 *            For a character driver, this field has two sub-fields          *
 *            . Bits b15..8  are the device number                           *
 *            . Bits b7..0   are the sub-channel number, 0xFF for the device *
 *****************************************************************************/

    mincode = pmd->rtemincode        ; /* Extract the route "mincode"        */
    maxcode = pmd->rtemaxcode        ; /* Extract the route "maxcode"        */
    minres  = -1                     ; /* With DRV_KEY_PRESS and             */
    maxres  = -1                     ; /* DRV_KEY_RELEASE, we accept all the */
                                       /* values or "reserve" (it is keycode)*/
    if (mincode EQ -1) goto step5    ; /* If no route has to be created then */
                                       /* skip the "add_uroute" call         */
    if (mincode EQ IND_REPORT)         /* With IND-REPORT, "reserve" is the  */
      BEGIN                            /* device or sub-channel IOD          */
        major  = iolib_modd[moidx]     /* Extract frol the MODD the major    */
               & 0x00FF0000          ; /* number and keep it in b23..16      */
        minres = major | 0x01000000  ; /* This interval will match all the   */
        maxres = major | 0x0100FFFF  ; /* IOD's for the driver, all the      */
      END_IF                           /* minor and sub-channels             */
                                       /*                                    */
    add_uroute(mincode            ,    /* Min value for "code"               */
               maxcode            ,    /* Max value for "code"               */
               minres             ,    /* Min value for "reserve"            */
               maxres             ,    /* Max value for "reserve"            */
               0                  ,    /* Flags                              */
               iolib_rteid + moidx ) ; /* Returned Route identifier          */


/*****************************************************************************
 * Step 5 : The VMIO driver is ready. We are going now to allocate the       *
 * ------   device. We call the "drv_alloc_device" procedure, this procedure *
 *          sends a REQ_ALLOC event to the VMIO driver FSM. This automaton   *
 *          will then send us a RESP_ALLOC response event. Here we           *
 *          unschedule until this event is received. We keep the returned    *
 *          IOD value in "iolob_iod[ioidx]" and we increment the count       *
 *          "iodnb[moidx]" of allocated devices for the driver               *
 *          ---------------------------------------------------------------- *
 *          We remind that bits b8..15 of any device IOD is the device       *
 *          number, starting with 0. So here, if there is no error, we get   *
 *          "device" from bits b8..15 of "iod"                               *
 *****************************************************************************/

    step5                            : /* Allocate the device                */
                                       /*                                    */
    hsprintf(name,"%s\\%s",            /* The device name is                 */
             modname      ,            /* "DRIVER_NAME\DEV" whereas "dev"    */
             devname        )        ; /* is the device name                 */
                                       /*                                    */
    r = drv_alloc_device(              /* Allocate device "name"             */
                 iolib_modd[moidx],    /* Module Descriptor Id               */
                 name             ,    /* Device name = name                 */
                 ""               ,    /* Tagini  = empty                    */
                 ""               ,    /* Taglist = empty                    */
                 pmd->flags       ,    /* Flags   = 0 or NBLOCKING_IO        */
                 2                ,    /* Reqid                              */
                 0                ,    /* Route ID   = empty                 */
                 0                ,    /* Route ID 2 = empty                 */
                 iod               ) ; /* Returned I/O descriptor            */
                                       /*                                    */
    if ( r ) goto exit               ; /* Exit if error                      */
                                       /*                                    */
    iolib_iod [ioidx] = *iod         ; /* Keep the allocated IOD value       */
    iolib_iodnb[moidx] ++            ; /* Update count of allocated devices  */


/*****************************************************************************
 * Step 6 : Here we initialize the hardware of the device. We call the       *
 * ------   "device_open_device" procedure. This procedures sends a REQ_OPEN *
 *          request event to the VMIO driver automaton, that is located      *
 *          inside the IO container. Then we ask to be unscheduled until the *
 *          AUT_LED send us the RESP_OPEN response event                     *
 *****************************************************************************/

    r = drv_open_device(               /* Sends a REQ_OPEN                   */
                    *iod          ,    /* I/O descriptor                     */
                    (CHAR*)taglist,    /* Taglist                            */
                    3              ) ; /* Reqid                              */

    exit                             : /* Exit label                         */
                                       /*                                    */

#ifdef TEST                            /* Begin conditional compilation      */
    procfin("io_open")               ; /* Message                            */
#endif                                 /* End   conditional compilation      */

    return( r )                      ; /* Return code of the procedure       */

  END_PROC                             /* End of the procedure               */


/*  Procedure io_close ------------------------------------------------------*/
/*
    Purpose : Closes and then frees a device. If all the devices of the driver
              are closed, then terminates the driver.

    Author : JYA                                 Last modification : 19/18/2022

...............................................................................

    Calling example : io_close( iod )

    Input     : iod      INT        Opened device IO descriptor

    Output    : 

    Notes : 

...............................................................................
*/

IWORD  PROCEDURE io_close(int iod)

  BEGIN                                /* Begin of this procedure            */
 
/*  Declaration of local variables of this procedure ........................*/

    IWORD          r                 ; /* Return code of the procedure       */
    INT            moidx             ; /* Indew in iodlib_modd / iolib_iodnb */
    INT            ioidx             ; /* Index in "iodlib_iod[]"            */
    INT            maj               ; /* Major number << 16                 */
    UINT           list[2]           ; /* Needed for "del_uroute"            */

/*  Beginning of the code ...................................................*/

#ifdef TEST                            /* Begin conditional compilation      */
    procdeb("io_close")              ; /* Message                            */
#endif                                 /* End   conditional compilation      */

/*****************************************************************************
 * Step 1 : Check that "iod" is a correct value, that it is a value          *
 * ------   previously returned by the "io_open" procedure. We do not allow  *
 *          a call to the present procedure with an IOD which hs not been    *
 *          allocated using "io_open". We call the "io_chk_iod" subroutine   *
 *****************************************************************************/

    r = io_chk_iod(iod,0,              /* Check that "iod" is one            */
                   &moidx,&ioidx)    ; /* "drv_alloc_device" returned value  */
                                       /*                                    */
    if ( r )                           /* Exit if "iod" is incorrect with    */
      goto exit                      ; /* the BAD_IOD error code             */


/*****************************************************************************
 * Step  2 : Search for the corresponding MODD in the "iolib_modd[]" table   *
 * -------   and set "moidx" to the index the MODD has been found. We simply *
 *           do a simple relational filtering because bits b16..23 of IOD's  *
 *           and MODD's are the major number, so:                            *
 *           - We set "maj" to bitsb16..23 of "iod", we extract the major    *
 *             number from the IOD.                                          *
 *           - Then we loop on entries of the "iolib_modd[]" table of MODD's,*
 *             using "moidx" as the index, and we compare bits b 16..23 of   *
 *             "iolib_modd[moidx]" with "maj". We stop looping when we       *
 *             have found our MODD, we jump to next step 3                   *
 *           --------------------------------------------------------------- *
 *           We remind that with drivers, we have at mod one MODD for every  *
 *           driver and one distinct major number per driver. As a           *
 *           consequence, the "iolib_modd[]" table cannot hold more than     *
 *           one MODD for a given major number                               *
 *           --------------------------------------------------------------- *
 *           Unless we have incoherent data (a bug) the search should never  *
 *           fail, we always have one MODD (and only one) for every IOD      *
 *****************************************************************************/

    r   = BAD_IOD                    ; /* Let'a assume we will fail          */
    maj = iod AND 0x00FF0000         ; /* Get the major number field value   */
                                       /*                                    */
    for(moidx = 0             ;        /* We loop with "moidx", searching    */
        moidx LT iolib_nbmodd ;        /* the "iodlib_modd[]" table for the  */
        moidx++            )           /* MODD which holds in bits b16..23   */
      BEGIN                            /* the major number for "iod".        */
        if ( (iolib_modd[moidx] AND    /* If this MODD holds the "maj" major */
              0x00FF0000) EQ maj   )   /* number, we are happy, we stop      */
          goto step3                 ; /* looping and jump to next step 3    */
      END_FOR                          /*                                    */
                                       /* The search failed: we have a bug   */
    goto exit                        ; /* since this should never happens    */


/*****************************************************************************
 * Step 3 : Close the device. We have to call "drv_close_device" for "iod".  *
 * -------  This procedure sends a REQ_CLOSE request event to the driver     *
 *          VMIO automaton and here we unschedule until the corresponding    *
 *          RESP_CLOSE, sent by AUT_xyz, is received                         *
 *****************************************************************************/

    step3                            : /* Close the device                   */
                                       /*                                    */
    r = drv_close_device(iod         , /* IOD of the device to be closed     */
                         4        )  ; /* Reqid                              */
    if ( r )                           /* If we have any error, we exit      */
      goto exit                      ; /* with that error code value         */


/*****************************************************************************
 * Step 4 : Free the device. We have to send to the AUT_xyz VMIO automaton a *
 * ------   REQ_FREE request event. We have to call the "drv_free" procedure *
 *          to do that. Then, we unschedule until the RESP_FREE response     *
 *          event is received.                                               *
 *****************************************************************************/

    r = drv_free(iod,                  /* IOD of the device to be freed      */
                 5   )               ; /* reqid                              */
    if ( r )                           /* This call should never fail        */
      goto exit                      ; /*                                    */


/*****************************************************************************
 * Step 5 :  We free the "ioidx" entry of the "iolib_iod[]" table and then   *
 * -------   decrement the count of allocated devices for that module, the   *
 *           "iolib_iodnb[moidx]" global variable. If this count is still    *
 *           strictly positive, if there is one or more IOD's, we are done   *
 *           and we exirt with "r=0" else we continue with next step 6.      *
 *****************************************************************************/

    iolib_iod  [ioidx] = 0           ; /* Free the "ioidx" entry of iolob_iod*/
    iolib_iodnb[moidx] --            ; /* Update the count of allocated      */
                                       /* devices/IOD's for the driver       */
                                       /*                                    */
    if (iolib_iodnb[moidx] GT 0)       /* If we still have one allocated     */
      goto exit                      ; /* device, we are done                */


/*****************************************************************************
 * Step 6 : The count of remaining allocated devices has reached 0, so       *
 * ------   we terminate the driver, we call the "drv_end" interface         *
 *          procedure that sends a REQ_END request event to the driver's     *
 *          VMIO  autiomaton and unschedule us until the RESP_END response   *
 *          event is received.                                               *
 *****************************************************************************/

    r = drv_end(iolib_modd[moidx] ,     /* We terminate the use of this      */
                6                  )  ; /* driver, we call "drv_end"         */
    if ( r )                            /*                                   */
      goto exit                       ; /* This should never occur           */


/*****************************************************************************
 * Step 7 : Delete the route we were using for IND_REPORT or DRV_KEY events. *
 * -------  We have to call the "del_uroute" procedure. As this procedure    *
 *          takes as in put an array containing a list of route identifiers, *
 *          we use  here an array of one entry in which we store the         *
 *          "iolib_rteid[moidx]" route identifier before calling  the        *
 *          "del_uroute" function. It has to be noted that we do not created *
 *          any route with the LED and DAC driver and in such a case         *
 *          "iolib_rteid[moidx]" is 0, nothing has to be done.               *
 *          ---------------------------------------------------------------- *
 *          We could directly call "del_uroute" with the address of the      *
 *          "iolib_rteid[moidx]" global variable because there is no         *
 *          difference between a variable adress and an array address        *
 *****************************************************************************/

    if ( iolib_rteid[moidx] EQ 0)      /* If there is no route to be deleted */
      goto exit                      ; /* we are done.                       */
                                       /*                                    */
    list[0] = iolib_rteid[moidx]     ; /* Store "rteid[moidx]" into a one    */
                                       /* entry array                        */
    del_uroute(list ,                  /* We call "del_uroute" with a list   */
               1     )               ; /* of route identifers which only     */
                                       /* contains one element.              */
    iolib_rteid[moidx] = 0           ; /* Clear the route identifier         */
                                       /*                                    */
    exit                             : /* Exit label                         */

#ifdef TEST                            /* Begin conditional compilation      */
    procfin("io_close")              ; /* Message                            */
#endif                                 /* End   conditional compilation      */

    return( r )                      ; /* Return code of the procedure       */

  END_PROC                             /* End of the procedure               */


/*  Procedure io_alloc_sub --------------------------------------------------*/
/*
    Purpose : Allocates one sub-channel of one opened device (GPIO or LED)

    Author : JYA                                 Last modification : 10/19/2022

...............................................................................

    Calling example : io_alloc_sub( iod, tagini, iodsub )

    Input     : iod      INT        Opened device IO descriptor   
                tagini   INT        Sub-channel selector tag list

    Output    : iodsub   INT*       Allocated sub-channel IO descriptor

    Notes : 

...............................................................................
*/

IWORD  PROCEDURE io_alloc_sub(int iod, const char *tagini, int *iodsub)

  BEGIN                                /* Begin of this procedure            */
 
/*  Declaration of local variables of this procedure ........................*/

    IWORD          r                 ; /* Return code of the procedure       */
    INT            moidx             ; /* Index in "iolib_modd[]"            */
    INT            ioidx             ; /* Index in "iolib_iod[]"             */
    INT            iodsubidx         ; /* Index in "iolib_iod[]"             */

/*  Beginning of the code ...................................................*/

#ifdef TEST                            /* Begin conditional compilation      */
    procdeb("io_alloc_sub")          ; /* Message                            */
#endif                                 /* End   conditional compilation      */

/*****************************************************************************
 * Step 1 : Check that "iod" is a correct value, that it is a value          *
 * ------   previously returned by the "io_open" procedure. We do not allow  *
 *          a call to the present procedure with an IOD which hs not been    *
 *          allocated using "io_open". We call the "io_chk_iod" subroutine   *
 *****************************************************************************/

    r = io_chk_iod(iod,0,              /* Check that "iod" is one            */
                   &moidx,&ioidx)    ; /* "drv_alloc_device" returned value  */
                                       /*                                    */
    if ( r )                           /* Exit if "iod" is incorrect with    */
      goto exit                      ; /* the BAD_IOD error code             */

/*****************************************************************************
 * Step  2 : Check we have at least one free entry in the "iolib_iod[]"      *
 * -------   table, because we will need to store the "*iodsub" IOD value.   *
 *           We use the "io_get_iodidx" subroutine, which returns            *
 *           NOT_ENOUGH_MEM if table is full.                                *
 *****************************************************************************/

    r = io_get_iodidx( &iodsubidx )  ; /* Check if "iolib_iod[]" is full,    */
                                       /* and return first free index        */
    if ( r )                           /* If the table is full, we exit      */
      goto exit                      ; /* with NOT_ENOUGH_MEM                */

/*****************************************************************************
 * Step 3 : We are going to allocate one pin/led subchannel. We call the     *
 * -------  "drv_alloc_subchan" procedure, this procedure sends a            *
 *          REQ_ALLOC_SUB event to the driver's VMIO automaton. This         *
 *          automaton will then send us a RESP_ALLOC_SUB response event.     *
 *          Here we unschedule (in "drv_alloc_subchan") until this event is  *
 *          received.                                                        *
 *****************************************************************************/

    r = drv_alloc_subchan(             /* Send a REQ_ALLOC_SUB               */
                     iod           ,   /* I/O descriptor of controller       */
                     (CHAR*)tagini ,   /* Tagini  = PIN/LED selector         */
                     ""            ,   /* Taglist = empty                    */
                     0             ,   /* Flags                              */
                     7             ,   /* Reqid                              */
                     0             ,   /* Route ID   = empty                 */
                     0             ,   /* Route ID 2 = empty                 */
                     iodsub        ) ; /* Returned I/O descriptor            */
                                       /*                                    */
    if ( r )                           /*                                    */
      goto exit                      ; /*                                    */

/*****************************************************************************
 * Step 4 : We are successful, keep the "*iodsub" IOD value in the           *
 * ------   "iolib_iod[iodsubidx]"                                           *
 *****************************************************************************/

    iolib_iod[iodsubidx] = *iodsub   ; /* Keep the allocated sub-channel IOD */
                                       /* at index "iodsubidx" of "iolib_iod"*/

    exit                             :

#ifdef TEST                            /* Begin conditional compilation      */
    procfin("io_alloc_sub")          ; /* Message                            */
#endif                                 /* End   conditional compilation      */

    return( r )                      ; /* Return code of the procedure       */

  END_PROC                             /* End of the procedure               */


/*  Procedure io_free_sub ---------------------------------------------------*/
/*
    Purpose : Free one sub-channel of one device. The sub-channel must have
              been allocated using "io_alloc_sub"

    Author : JYA                                 Last modification : 10/19/2022

...............................................................................

    Calling example : io_free_sub ( iodsub )

    Input     : iodsub   int        Sub-channel IO descriptor

    Output    : valproc  IWORD      Function return/error code

    Notes : 

...............................................................................
*/

IWORD  PROCEDURE io_free_sub(int iodsub)

  BEGIN                                /* Begin of this procedure            */
 
/*  Declaration of local variables of this procedure ........................*/

    IWORD          r                 ; /* Return code of the procedure       */
    INT            moidx             ; /* Indew in iodlib_modd / iolib_iodnb */
    INT            ioidx             ; /* Index in "iodlib_iod[]"            */

/*  Beginning of the code ...................................................*/

#ifdef TEST                            /* Begin conditional compilation      */
    procdeb("io_free_sub")           ; /* Message                            */
#endif                                 /* End   conditional compilation      */

/*****************************************************************************
 * Step 1 : Check that "iod" is a correct value, that it is a value          *
 * ------   previously returned by the "io_open" procedure. We do not allow  *
 *          a call to the present procedure with an IOD which hs not been    *
 *          allocated using "io_open". We call the "io_chk_iod" subroutine   *
 *****************************************************************************/

    r = io_chk_iod(iodsub , 1,         /* Check that "iod" is one            */
                   &moidx,&ioidx)    ; /* "drv_alloc_device" returned value  */
                                       /*                                    */
    if ( r )                           /* Exit if "iod" is incorrect with    */
      goto exit                      ; /* the BAD_IOD error code             */


/*****************************************************************************
 * Step 2 : Free the subchannel. We call "drv_free_subchan", this procedure  *
 * ------   sends a REQ_FREE_SUB request event to the driver's VMIO automaton*
 *          and here we unschedule until the corresponding RESP_FREE_SUB,    *
 *          sent by the driver's VMIO FSM is received.                       *
 *****************************************************************************/

    r = drv_free_subchan(iodsub ,      /* IOD of the subchannel to be freed  */
                         8       )   ; /* Reqid                              */
    if ( r )                           /* If we have any error, we exit      */
      goto exit                      ; /* with that error code value         */


/*****************************************************************************
 * Step 3 : Fre the "ioidx" entry of table "iolib_iod[]". We wimply need     *
 * -------  to clear that value                                              *
 *****************************************************************************/

    iolib_iod[ioidx] = 0             ; /* Clear/free entry "ioidx"           */

    exit                             : /* Procedure exit label               */
                                       /*                                    */
#ifdef TEST                            /* Begin conditional compilation      */
    procfin("io_free_sub")           ; /* Message                            */
#endif                                 /* End   conditional compilation      */

    return( r )                      ; /* Return code of the procedure       */

  END_PROC                             /* End of the procedure               */


/*  Procedure io_write ------------------------------------------------------*/
/*
    Purpose : Writes one data buffer. This is a blocking call, caller is
              unscheduled until the end of sending/write, despite the fact
              the driver has been opened in NBLOCKING_IO

    Author : JYA                                 Last modification : 11/05/2022

...............................................................................

    Calling example : 

    Input     : x        IWORD      C

    Output    : 

    Notes : 

...............................................................................
*/

IWORD  PROCEDURE io_write(int iod, unsigned char *buf, int lg,int flags)

  BEGIN                                /* Begin of this procedure            */
 
/*  Declaration of local variables of this procedure ........................*/

    IWORD          r                 ; /* Return code of the procedure       */
    INT            moidx             ; /* Indew in iodlib_modd / iolib_iodnb */
    INT            ioidx             ; /* Index in "iodlib_iod[]"            */
    INT            lgmax             ; /* Max data bytes count               */
    INT            offset, trail, len; /* Returned by "drv_get_buf_size"     */
    UINT           waitlist[1][3]    ; /* Parameter of "waitevt_task         */

/*  Beginning of the code ...................................................*/

#ifdef TEST                            /* Begin conditional compilation      */
    procdeb("io_write")              ; /* Message                            */
#endif                                 /* End   conditional compilation      */

/*****************************************************************************
 * Step 1 : Check that "iod" is a correct value, that it is a value          *
 * ------   previously returned by the "io_open" procedure. We do not allow  *
 *          a call to the present procedure with an IOD which hs not been    *
 *          allocated using "io_open". We call the "io_chk_iod" subroutine   *
 *****************************************************************************/

    r = io_chk_iod(iod, 2,             /* Check that "iod" is one            */
                   &moidx,&ioidx)    ; /* "drv_alloc_device" returned value  */
                                       /*                                    */
    if ( r )                           /* Exit if "iod" is incorrect with    */
      goto exit                      ; /* the BAD_IOD error code             */


/*****************************************************************************
 * Step 2 : With a protocol, a header in the sending buffer may be needed    *
 * ------   The "drv_get_bufsize" procedure returns the offset where the     *
 *          first user data byte should be located. We assume that the       *
 *          caller of "io_write" did the thing right, the he took in account *
 *          this same offset correctly, wrote his first data byte at "offset"*
 *          The "drv_write" procedure checks that the "offset" it gets has   *
 *          exactly the correct value otherwise BAD_PARAMETER is returned    *
 *          Then we check that our "lg" count of user data bytes is not      *
 *          greater than the "lgmax" limit, unles we exit with a             *
 *          BAD_PARAMETER error code                                         *
 *****************************************************************************/

    r = drv_get_bufsize( iod     ,     /* iod    IO descriptor               */
                         lg      ,     /* lg     Count of user bytes         */
                         0       ,     /* wrrd   0:write 1:read              */
                         &lgmax  ,     /* lgmax  Max data byte count         */
                         &offset ,     /* offset Offset to 1st byte to send  */
                         &trail  ,     /* trail  Trailer sier in bytes       */
                         &len      ) ; /* len    "offset + lg + trail"       */
    if ( r )                           /* If we fail, we exit, returning the */
      goto exit                      ; /* "drv_get_bufsize" error code value */
                                       /*                                    */
    if ( lg GT lgmax )                 /* Check the count of user data byte  */
      BEGIN                            /* "lg" is not greater than the       */
        r = BAD_PARAMETER            ; /* "lgmax" limit, unless we exit with */
        goto exit                    ; /* a BAD_PARAMETER error code         */
      END_IF                           /*                                    */


/*****************************************************************************
 * Step 3 : Send using "iod" the "lg" bytes that are in the "buf" data       *
 * ------   buffer.  We call the "drv_write" procedure that sends a          *
 *          REQ_WRITE request event to the driver's VMIO automaton.          *
 *****************************************************************************/

    r = drv_write(iod       ,          /* IOD of serial controller DEV0      */
                  buf       ,          /* Buffer Address                     */
                  offset    ,          /* Buffer Offset                      */
                  lg        ,          /* Number of bytes to be sent         */
                  flags     ,          /* Flags                              */
                  12          )      ; /* Reqid                              */
    if ( r )                           /* If we fail, we exit, returning the */
      goto exit                      ; /* "drv_write" error code value       */
                                       /*                                    */

/*****************************************************************************
 * Step 4 : We unschedule until the RESP_WRITE event is received. This event *
 * ------   is sent by the VMIO automaton only when the last byte as been    *
 *          sent and not before. We build a wait list containing a single    *
 *          pair (code=RESP_WRITE,reserve=iod). The task will be unschedule  *
 *          until an event corresponding to the one expected is received by  *
 *          the task. This wait may be infinite if the RESP_WRITE event is   *
 *          never received.                                                  *
 *****************************************************************************/

    waitlist[0][0]  = WAIT_CODERES   ; /* Test "code" and "reserve" fields   */
    waitlist[0][1]  = RESP_WRITE     ; /* The awaited value for "code"       */
    waitlist[0][2]  = iod            ; /* The awaited value for "reserve"    */
                                       /*                                    */
    r = waitevt_task(waitlist ,        /* Address of waiting list            */
                     1        ,        /* Size of "waitlist[]"               */
                     0        ,        /* No maximum waiting time            */
                     0          )    ; /* Do not purge previous events       */
                                       /*                                    */
    if ( r ) goto exit               ; /* If wae had an error, exit          */

    r = task_evt.length              ; /* Count of sent bytes or error       */

/*****************************************************************************
 * Step 5 : If we have any error, we call the OS halt "vhalt" function       *
 * ------                                                                    *
 *****************************************************************************/

    exit                             : /* Exit label                         */
                                       /*                                    */
    if ( r LT 0 )                      /* If we had any error                */
      vhalt(0)                       ; /* call the OS halt function          */

#ifdef TEST                            /* Begin conditional compilation      */
    procfin("io_write")              ; /* Message                            */
#endif                                 /* End   conditional compilation      */

    return( r )                      ; /* Return code of the procedure       */

  END_PROC                             /* End of the procedure               */


/*  Procedure io_send -------------------------------------------------------*/
/*
    Purpose : Request the sending of one data buffer to the driver. This
              function is non blocking.

    Author : JYA                                 Last modification : 11/12/2022

...............................................................................

    Calling example : io_send( iod, buf, lg , count )

    Input     : x        IWORD      C

    Output    : 

    Notes : 

...............................................................................
*/

IWORD  PROCEDURE io_send(int iod, unsigned char *buf, int lg, int *count)

  BEGIN                                /* Begin of this procedure            */
 
/*  Declaration of local variables of this procedure ........................*/

    IWORD          r                 ; /* Return code of the procedure       */
    INT            moidx             ; /* Indew in iodlib_modd / iolib_iodnb */
    INT            ioidx             ; /* Index in "iodlib_iod[]"            */
    INT            lgmax             ; /* Max count of data bytes            */
    INT            offset,trail,len  ; /* drv_get_bufsize returned values    */

/*  Beginning of the code ...................................................*/

#ifdef TEST                            /* Begin conditional compilation      */
    procdeb("io_send")               ; /* Message                            */
#endif                                 /* End   conditional compilation      */

/*****************************************************************************
 * Step 1 : Check that "iod" is a correct value, that it is a value          *
 * ------   previously returned by the "io_open" procedure. We do not allow  *
 *          a call to the present procedure with an IOD which hs not been    *
 *          allocated using "io_open". We call the "io_chk_iod" subroutine   *
 *****************************************************************************/

    r = io_chk_iod(iod, 2,             /* Check that "iod" is one            */
                   &moidx,&ioidx)    ; /* "drv_alloc_device" returned value  */
                                       /*                                    */
    if ( r )                           /* Exit if "iod" is incorrect with    */
      goto exit                      ; /* the BAD_IOD error code             */


/*****************************************************************************
 * Step 2 : Retreive the "offset" parameter value for "iod". This value is   *
 * ------   0 for ASY/GPIO and strictly positive with CAN. We call the       *
 *          "drv_get_bufsize" procedure to get "offset". The function also   *
 *          returns the "lgmax" maximum count of user data bytes and we      *
 *          check that "lg" is not greater than "lgmax" unless we exit with  *
 *          a BAD_PARAMETER error code value                                 *
 *****************************************************************************/

    r = drv_get_bufsize(iod     ,      /* Unique identifier of channel       */
                        lg      ,      /* Count of user data bytes           */
                        0       ,      /* 0:write  1:read                    */
                        &lgmax  ,      /* Max count of data bytes            */
                        &offset ,      /* Offset of 1st byte to be sent      */
                        &trail  ,      /* Trailer size in bytes              */
                        &len     )   ; /* "offet" + "lg" + "trail"           */
                                       /*                                    */
    if ( r )                           /* If any error occured, we exit with */
      goto exit                      ; /* "drv_get_bufsize" error code       */
                                       /*                                    */
    if ( lg GT lgmax )                 /* Check the count of user data byte  */
      BEGIN                            /* "lg" is not greater than the       */
        r = BAD_PARAMETER            ; /* "lgmax" limit, unless we exit with */
        goto exit                    ; /* a BAD_PARAMETER error code         */
      END_IF                           /*                                    */

/*****************************************************************************
 * Step 3 : Send using "iod" the "lg" bytes that are in telecom buffer       *
 * ------   pointed by "buf".  We call the "drv_write" procedure that sends  *
 *          a REQ_WRITE request event to the driver VMIO automaton. We want  *
 *          to maximize throughput, so we don't wait for data ta have been   *
 *          sent here. We therefore do not unschedule.                       *
 *****************************************************************************/

    r = drv_write(iod       ,          /* IOD of serial controller DEV0      */
                  buf       ,          /* Buffer Address                     */
                  offset    ,          /* Buffer Offset                      */
                  lg        ,          /* Number of bytes to be sent         */
                  0         ,          /* Flags                              */
                  13          )      ; /* Reqid                              */
                                       /*                                    */
    *count += 1                      ; /* Update number of inflight messages */


/*****************************************************************************
 * Step 4 : If an error occured, call the "vhalt()" OS halt function         *
 *****************************************************************************/

    exit                             : /* Procedure exit label               */
                                       /*                                    */
    if ( r )                           /* If any error occured, call the     */
      vhalt(0)                       ; /* OS abort function                  */

#ifdef TEST                            /* Begin conditional compilation      */
    procfin("io_send")               ; /* Message                            */
#endif                                 /* End   conditional compilation      */

    return( r )                      ; /* Return code of the procedure       */

  END_PROC                             /* End of the procedure               */


/*  Procedure io_read -------------------------------------------------------*/
/*
    Purpose : Blocking read from a device

    Author : JYA                                 Last modification : 10/23/2022

...............................................................................

    Calling example : ioread

    Input     : x        IWORD      C

    Output    : 

    Notes : 

...............................................................................
*/

IWORD  PROCEDURE io_read(int iod, unsigned char *buf, int size, int *len)

  BEGIN                                /* Begin of this procedure            */
 
/*  Declaration of local variables of this procedure ........................*/

    IWORD          r                 ; /* Return code of the procedure       */
    INT            moidx             ; /* Indew in iodlib_modd / iolib_iodnb */
    INT            ioidx             ; /* Index in "iodlib_iod[]"            */

/*  Beginning of the code ...................................................*/

#ifdef TEST                            /* Begin conditional compilation      */
    procdeb("io_read")               ; /* Message                            */
#endif                                 /* End   conditional compilation      */

/*****************************************************************************
 * Step 1 : Check that "iod" is a correct value, that it is a value          *
 * ------   previously returned by the "io_open" procedure. We do not allow  *
 *          a call to the present procedure with an IOD which hs not been    *
 *          allocated using "io_open". We call the "io_chk_iod" subroutine   *
 *****************************************************************************/

    r = io_chk_iod(iod, 2,             /* Check that "iod" is one            */
                   &moidx,&ioidx)    ; /* "drv_alloc_device" returned value  */
                                       /*                                    */
    if ( r )                           /* Exit if "iod" is incorrect with    */
      goto exit                      ; /* the BAD_IOD error code             */

/*****************************************************************************
 * Step  x : .                                                               *
 * -------   .                                                               *
 *           .                                                               *
 *****************************************************************************/

    ;                                  /*                                    */
    ;                                  /*                                    */
    ;                                  /*                                    */

    exit                             : /* Exit label                         */

#ifdef TEST                            /* Begin conditional compilation      */
    procfin("io_read")               ; /* Message                            */
#endif                                 /* End   conditional compilation      */

    return( r )                      ; /* Return code of the procedure       */

  END_PROC                             /* End of the procedure               */

/*  Procedure io_receive ----------------------------------------------------*/
/*
    Purpose : Gives receive tokens to a driver

    Author : JYA                                 Last modification : 10/19/2022

...............................................................................

    Calling example : io_receive ( iod, nbtok, size, count )

    Input     : iod      INT        Device or sub-channel IO descriptor

    Output    : 

    Notes : 

...............................................................................
*/

IWORD  PROCEDURE io_receive( int iod, int nbtok, int size, int *count )

  BEGIN                                /* Begin of this procedure            */
 
/*  Declaration of local variables of this procedure ........................*/

    IWORD          r                 ; /* Return code of the procedure       */
    INT            moidx             ; /* Indew in iodlib_modd / iolib_iodnb */
    INT            ioidx             ; /* Index in "iodlib_iod[]"            */
    IWORD          i                 ; /* Loop counter                       */
    UCHAR          *adr              ; /* For "drv_read"                     */
    INT            len, offset, lg   ; /* Returned by "drv_read"             */

/*  Beginning of the code ...................................................*/

#ifdef TEST                            /* Begin conditional compilation      */
    procdeb("io_receive")            ; /* Message                            */
#endif                                 /* End   conditional compilation      */

/*****************************************************************************
 * Step 1 : Check that "iod" is a correct value, that it is a value          *
 * ------   previously returned by the "io_open" procedure. We do not allow  *
 *          a call to the present procedure with an IOD which hs not been    *
 *          allocated using "io_open". We call the "io_chk_iod" subroutine   *
 *****************************************************************************/

    r = io_chk_iod(iod, 2,             /* Check that "iod" is one            */
                   &moidx,&ioidx)    ; /* "drv_alloc_device" returned value  */
                                       /*                                    */
    if ( r )                           /* Exit if "iod" is incorrect with    */
      goto exit                      ; /* the BAD_IOD error code             */


/*****************************************************************************
 * Step 2 : In order to have data flow control, the application have to      *
 * ------   give to the driver receive tokens. One token is one buffer.      *
 *          The applications gives to the driver a number of tokens "nbtok"  *
 *          and the size "size" of the buffers. The application as to send   *
 *          a REQ_READ request to the VMIO automaton. To allocate the "nbtok"*
 *          buffers and send one REQ_READ, we call "drv_read".               *
 *          ---------------------------------------------------------------- *
 *          RESP_READ event will be received after data will be received     *
 *          from the telecomunication line. This may take a long time, or    *
 *          may never occur. So we are not going to unschedule right now     *
 *****************************************************************************/

    adr = HNULL                      ; /* No buffer address, it will be      */
                                       /* allocated by "drv_read"            */
    for (i = 0 ; i LT nbtok ; i++)     /* Loop with the "nbtok" read tokens  */
      BEGIN                            /* to be given to the driver          */
        r = drv_read(iod       ,       /* I/O descriptor for DEV0            */
                     size      ,       /* Maximum number of bytes / buffer   */
                     0         ,       /* Flags                              */
                     &adr      ,       /* Buffer address                     */
                     &len      ,       /* Allocated buffer address           */
                     &offset   ,       /* Offset (unused)                    */
                     &lg       ,       /* Number of bytes read (unused)      */
                     11         )    ; /* Reqid                              */
        if ( r )                       /* If operation failed,               */
          break                      ; /* we stop looping                    */
                                       /*                                    */
        *count += 1                  ; /* Increment the count of tokens      */
      END_FOR

/*****************************************************************************
 * Step 3 : If we have any error, we call the OS halt "vhalt" function       *
 * ------                                                                    *
 *****************************************************************************/

    exit                             : /* Exit label                         */
    if ( r )                           /* If an error has occured            */
      vhalt(0)                       ; /* OS stop if error                   */


#ifdef TEST                            /* Begin conditional compilation      */
    procfin("io_receive")            ; /* Message                            */
#endif                                 /* End   conditional compilation      */

    return( r )                      ; /* Return code of the procedure       */

  END_PROC                             /* End of the procedure               */


/*  Procedure io_setval -----------------------------------------------------*/
/*
    Purpose : Sets/writes a taglist

    Author : JYA                                 Last modification : 10/22/2022

...............................................................................

    Calling example : io_setval( iod, taglist )

    Input     : iod      IWORD        Device or sub-channel reference
                taglist  const char*  List of tags

    Output    : 

    Notes : 

...............................................................................
*/

IWORD  PROCEDURE io_setval(int iod, const char *taglist)

  BEGIN                                /* Begin of this procedure            */
 
/*  Declaration of local variables of this procedure ........................*/

    IWORD          r                 ; /* Return code of the procedure       */
    INT            moidx             ; /* Indew in iodlib_modd / iolib_iodnb */
    INT            ioidx             ; /* Index in "iodlib_iod[]"            */

/*  Beginning of the code ...................................................*/

#ifdef TEST                            /* Begin conditional compilation      */
    procdeb("io_setval")             ; /* Message                            */
#endif                                 /* End   conditional compilation      */

/*****************************************************************************
 * Step 1 : Check that "iod" is a correct value, that it is a value          *
 * ------   previously returned by the "io_open" procedure. We do not allow  *
 *          a call to the present procedure with an IOD which hs not been    *
 *          allocated using "io_open". We call the "io_chk_iod" subroutine   *
 *****************************************************************************/

    r = io_chk_iod(iod, 2,             /* Check that "iod" is one            */
                   &moidx,&ioidx)    ; /* "drv_alloc_device" returned value  */
                                       /*                                    */
    if ( r )                           /* Exit if "iod" is incorrect with    */
      goto exit                      ; /* the BAD_IOD error code             */


/*****************************************************************************
 * Step 2 : Write the taglist "taglist" to the driver. The "drv_setval"      *
 * ------   procedure checks the syntax of the "taglist" string and if there *
 *          is an error it immediately returns with an error code. If the    *
 *          syntax is correct, a telecom buffer is allocated. If this        *
 *          allocation fails, the procedure also returns immediately with an *
 *          error code. If the buffer allocation is a success, the           *
 *          compiled register list is put in the telecom buffer, that is     *
 *          pointed by a REQ_SETVAL event that is sent to the driver's VMIO  *
 *          automaton. The "drv_setval" unschedules our task, until the      *
 *          REST_SETVAL response event is received from the driver's FSM.    *
 *****************************************************************************/

    r = drv_setval(iod             ,   /* I/O descriptor for DEV1            */
                   (CHAR*)taglist  ,   /* taglist                            */
                   10               ); /* Reqid                              */
                                       /*                                    */

/*****************************************************************************
 * Step 3 : If we have any error, we call the OS halt "vhalt" function       *
 * ------                                                                    *
 *****************************************************************************/

    exit                             : /* Procedure exit label               */
                                       /*                                    */
    if ( r )                           /* If we have any error, we call      */
      vhalt(0)                       ; /* the OS halt function               */

#ifdef TEST                            /* Begin conditional compilation      */
    procfin("io_setval")             ; /* Message                            */
#endif                                 /* End   conditional compilation      */

    return( r )                      ; /* Return code of the procedure       */

  END_PROC                             /* End of the procedure               */


/*  Procedure io_getval -----------------------------------------------------*/
/*
    Purpose : Reads, get a tag list.

    Author : JYA                                 Last modification : 10/19/2022

...............................................................................

    Calling example : io_getval( iod, taglist, sz )

    Input     : iod      int        Device or sub-channel IOD descriptor
                taglist  char*      Tags read buffer
                sz       int        Read buffer size in bytes

    Output    : valproc  IOWRD      Function return/error code

    Notes : 

...............................................................................
*/

IWORD  PROCEDURE io_getval(int iod, char *taglist, int sz)

  BEGIN                                /* Begin of this procedure            */
 
/*  Declaration of local variables of this procedure ........................*/

    IWORD          r                 ; /* Return code of the procedure       */
    INT            moidx             ; /* Indew in iodlib_modd / iolib_iodnb */
    INT            ioidx             ; /* Index in "iodlib_iod[]"            */

/*  Beginning of the code ...................................................*/

#ifdef TEST                            /* Begin conditional compilation      */
    procdeb("io_getval")             ; /* Message                            */
#endif                                 /* End   conditional compilation      */

/*****************************************************************************
 * Step 1 : Check that "iod" is a correct value, that it is a value          *
 * ------   previously returned by the "io_open" procedure. We do not allow  *
 *          a call to the present procedure with an IOD which hs not been    *
 *          allocated using "io_open". We call the "io_chk_iod" subroutine   *
 *****************************************************************************/

    r = io_chk_iod(iod, 2,             /* Check that "iod" is one            */
                   &moidx,&ioidx)    ; /* "drv_alloc_device" returned value  */
                                       /*                                    */
    if ( r )                           /* Exit if "iod" is incorrect with    */
      goto exit                      ; /* the BAD_IOD error code             */


/*****************************************************************************
 * Step 2 : Read the taglist "taglist" from the driver. The "drv_getval"     *
 * ------   procedure checks the syntax of the "taglist" template string and *
 *          if there is an error it immediately returns with an error code.  *
 *          If the syntax is correct, a telecom buffer is allocated. If this *
 *          allocation fails, the procedure also returns immediately with an *
 *          error code. If the buffer allocation is a success, the           *
 *          compiled register list is put in the telecom buffer, that is     *
 *          pointed by a REQ_GETVAL event that is sent to the driver's VMIO  *
 *          automaton. The VMIO FSM fills the register list with the tags    *
 *          values to be read and returns a RESP_GETVAL that points to the   *
 *          modified register list. Then the returned register list is       *
 *          converted to a character string. The "drv_getval" unschedules    *
 *          our task, until the REST_GETVAL response event is received from  *
 *          the driver's FSM.                                                *
 *****************************************************************************/

    r = drv_getval(iod      ,          /* I/O descriptor for DEV1            */
                   taglist  ,          /* taglist buffer                     */
                   sz       ,          /* taglist buffer sioze in bytes      */
                   9          )      ; /* Reqid                              */


/*****************************************************************************
 * Step 3 : If we have any error, we call the OS halt "vhalt" function       *
 * ------                                                                    *
 *****************************************************************************/

    exit                             : /* Procedure exit label               */
                                       /*                                    */
    if ( r )                           /* If we have any error, we call      */
      vhalt(0)                       ; /* the OS halt function               */

#ifdef TEST                            /* Begin conditional compilation      */
    procfin("io_getval")             ; /* Message                            */
#endif                                 /* End   conditional compilation      */

    return( r )                      ; /* Return code of the procedure       */

  END_PROC                             /* End of the procedure               */


/*  Procedure io_bufsize ----------------------------------------------------*/
/*
    Purpose : Returns the "len" neede buffer size in bytes in order to
              accomodate for "lg" data bytes.

    Author : JYA                                 Last modification : 10/19/2022

...............................................................................

    Calling example : 

    Input     : x        IWORD      C

    Output    : 

    Notes : 

...............................................................................
*/

IWORD  PROCEDURE io_bufsize(int iod, int lg, int *offset, int *len)

  BEGIN                                /* Begin of this procedure            */
 
/*  Declaration of local variables of this procedure ........................*/

    IWORD          r                 ; /* Return code of the procedure       */
    INT            moidx             ; /* Indew in iodlib_modd / iolib_iodnb */
    INT            ioidx             ; /* Index in "iodlib_iod[]"            */
    INT            lgmax             ; /* Unused value                       */
    INT            trail             ; /* Unused value                       */

/*  Beginning of the code ...................................................*/

#ifdef TEST                            /* Begin conditional compilation      */
    procdeb("io_bufsize")            ; /* Message                            */
#endif                                 /* End   conditional compilation      */

/*****************************************************************************
 * Step 1 : Check that "iod" is a correct value, that it is a value          *
 * ------   previously returned by the "io_open" procedure. We do not allow  *
 *          a call to the present procedure with an IOD which hs not been    *
 *          allocated using "io_open". We call the "io_chk_iod" subroutine   *
 *****************************************************************************/

    r = io_chk_iod(iod, 2,             /* Check that "iod" is one            */
                   &moidx,&ioidx)    ; /* "drv_alloc_device" returned value  */
                                       /*                                    */
    if ( r )                           /* Exit if "iod" is incorrect with    */
      goto exit                      ; /* the BAD_IOD error code             */


/*****************************************************************************
 * Step  2 : Call the "drv_get_bufsize" function for the "write" case        *
 * --------                                                                  *
 *****************************************************************************/

    r = drv_get_bufsize( iod    ,      /* Unique identifier of the channel   */
                         lg     ,      /* Count of user data bytes           */
                         0      ,      /* 0:write 1:read                     */
                         &lgmax ,      /* Maximum count of user data bytes   */
                         offset ,      /* Offset to the 1stt byte to send    */
                         &trail ,      /* Trailer size in bytes              */
                         len     )   ; /* Equal to "offset + lg + trail      */


    exit                             : /* Exit label                         */

#ifdef TEST                            /* Begin conditional compilation      */
    procfin("io_bufsize")            ; /* Message                            */
#endif                                 /* End   conditional compilation      */

    return( r )                      ; /* Return code of the procedure       */

  END_PROC                             /* End of the procedure               */


/*  Procedure io_get_iodidx -------------------------------------------------*/
/*
    Purpose : Subroutine which returns a free entry index in the "iolib_iod[]"
              table. The function simply scans the table and returns the
              first free entry index.

    Author : JYA                                 Last modification : 10/19/2022

...............................................................................

    Calling example : io_get_iodidx ( idx )

    Input     : -

    Output    : idx      INT*       Index in the "iolib_iod[]" table
                valproc  IWORD      0:Success NOT_ENOUGH_MEM: Table full

    Notes : 

...............................................................................
*/

static IWORD  PROCEDURE io_get_iodidx(INT *idx)

  BEGIN                                /* Begin of this procedure            */
 
/*  Declaration of local variables of this procedure ........................*/

    IWORD          r                 ; /* Return code of the procedure       */
    INT            ioidx             ; /* Index in the "iolib_iod[]" table   */

/*  Beginning of the code ...................................................*/

#ifdef TEST                            /* Begin conditional compilation      */
    procdeb("io_get_iodidx")         ; /* Message                            */
#endif                                 /* End   conditional compilation      */

/*****************************************************************************
 * Step  1 : Scan the "iolib_iod[]" table, looking for the first free        *
 * -------   entry, with a 0 value. We use "ioidx" as the current index      *
 *           value. Is no entry is found, we exit with NOT_ENOUGH_MEM        *
 *****************************************************************************/

    r = 0                            ; /* Let's assume success               */
                                       /*                                    */
    for (ioidx = 0            ;        /* We scan the "iolib_iod[]" array,   */
         ioidx LT iolib_nbiod ;        /* using "ioidx" as the loop and      */
         ioidx ++              )       /* array index                        */
      BEGIN                            /*                                    */
        if ( iolib_iod[ioidx] EQ 0 )   /* If entry "ioidx" is free, if the   */
          goto exit                  ; /* value is 0, we continue with next  */
      END_FOR                          /* step 2. Else we exit, returning    */
                                       /* the NOT_ENOUGH_MEM error code      */
    r = NOT_ENOUGH_MEM               ; /* Let's assume we will fail          */


/*****************************************************************************
 * Step  2 : We have to return the "idx" value. If we failed we return -1    *
 * --------  else we return the "ioidx" found index value                    *
 *****************************************************************************/

    exit                             : /* Procedure exit label               */
                                       /*                                    */
    *idx = r ? -1                      /* If we fail, we return -1           */
             : ioidx                 ; /* If successful, return ioidx        */

#ifdef TEST                            /* Begin conditional compilation      */
    procfin("io_get_iodidx")         ; /* Message                            */
#endif                                 /* End   conditional compilation      */

    return( r )                      ; /* Return code of the procedure       */

  END_PROC                             /* End of the procedure               */


/*  Procedure io_chk_driver -------------------------------------------------*/
/*
    Purpose : Check that a driver name is known to the present module. This
              module owns a dedicated table "iolib_moddef" which gives the list
              of driver the present module is allowed to use. This list is
              a subset of the list returned by "drv_list_module".

    Author : JYA                                 Last modification : 10/23/2022

...............................................................................

    Calling example : io_chk_driver ( driver, moidx, maj )

    Input     : driver   CHAR*      Driver name we want to check

    Output    : moidx    INT*       Found index in the "iolib_moddef[]" table
                maj      INT*       Foud index in the "vmos_driver" table
                valproc  IWORD      Error code

    Notes : 

...............................................................................
*/

static IWORD  PROCEDURE io_chk_driver(const CHAR *driver, INT *moidx, INT *maj)

  BEGIN                                /* Begin of this procedure            */
 
/*  Declaration of local variables of this procedure ........................*/

    IWORD             r              ; /* Return code of the procedure       */
    IWORD             i              ; /* Loop counter for "iolib_moddef[]"  */
    const Mod_def     *pmd           ; /* Pointer to "iolib_moddef[i]"       */
    IWORD             j              ; /* Loop counter for "vmos_driver[]"   */
    const Sys_driver  *vmd           ; /* Pointer to "vmos_driver[j]"        */

/*  Beginning of the code ...................................................*/

#ifdef TEST                            /* Begin conditional compilation      */
    procdeb("io_chk_driver")         ; /* Message                            */
#endif                                 /* End   conditional compilation      */

/*****************************************************************************
 * Step  1 : Search for the "driver" string in the "iolib_moddef[]" table. We*
 * -------   loop on the table entries, comparing "driver" with the          *
 *           "modname" fields of the "iolib_moddef" entries. If the "driver" *
 *           string is not found, we exit with BAD_NOMP. User gave us an     *
 *           unknown name.                                                   *
 *****************************************************************************/

    r      = BAD_NOMP                ; /* Let's assume we will fail          */
    *moidx = -1                      ; /* Invalid/empty values for both      */
    *maj   = -1                      ; /* "moidx" and "maj"                  */
                                       /*                                    */
    for(i = 0, pmd = iolib_moddef ;    /* We scan the "iolib_nbmodd" entries */
        i LT iolib_nbmodd         ;    /* of the "iolib_moddef[]" const table*/
        i++, pmd++            )        /*                                    */
      BEGIN                            /* If the "modname" string of the     */
        if (Strcmp(driver,             /* current "pmd" entry is equal to    */
                   pmd->modname) EQ 0) /* the "driver" string we are looking */
          goto step2                 ; /* for, we stop looping and continue  */
      END_FOR                          /* with next step 2                   */
                                       /*                                    */
    goto exit                        ; /* We failed. We exit with BAD_NOMP   */


/*****************************************************************************
 * Step  2 : Search for the "driver" string in the "vmos_driver[]" table. We *
 * -------   loop on the table entries, we compare "driver" with the "name"  *
 *           field of the "Sys_driver" structure. If the search fails, we    *
 *           have a major issue: The name is in the "iolib_moddef[]" table   *
 *           but not into "vmos_driver", implying the two tables are not     *
 *           coherent, "space_appcont_xyz.c" must be corrected. Indeed, we   *
 *           only take in account "vmos_driver[]" entries corresponding to   *
 *           character drivers, with "typdrv" EQ to DRV_CAR.                 *
 *****************************************************************************/

    step2                            : /* Search "driver" in "vmod_driver[]" */
                                       /*                                    */
    r = -1                           ; /* Let's assume we will fail          */
                                       /*                                    */
    for(j=0; j LT vmos_nbdriver; j++)  /* Scan the "vmos_driver[]", which    */
      BEGIN                            /* is a list of structure addresses   */
        vmd = vmos_driver[j]         ; /* Get the structure address # i      */
                                       /*                                    */
        if (vmd->typdrv NE DRV_CAR)    /* If the structure is not the one    */
          continue                   ; /* of a character driver, skip it     */
                                       /*                                    */
        if(Strcmp(driver,              /* If string "vmd->name" is the one   */
                  vmd->name) EQ 0)     /* we are looking for, we are done,   */
          goto step3                 ; /* we exit for the loop and continue  */
      END_FOR                          /* with next step 3                   */
                                       /*                                    */
    vhalt(0)                         ; /* We failed because of incorrect     */
    goto exit                        ; /* data in "space_appcont_xyz.c"      */


/*****************************************************************************
 * Step  3 : We are OK, the "driver" string has been found in both the       *
 * -------   "iolib_moddef[]" and the "vmos_driver[]" tables:                *
 *           - We clear the "r" function return code - No error              *
 *           - We set the "moidx" index in "iolib_moddef[]" to "i"           *
 *           - We set the "maj" index in "vmos_driver[]" to "j"              * 
 *****************************************************************************/

    step3                            : /* We are OK                          */
                                       /*                                    */
    r      = 0                       ; /* We clear the error code            */
    *moidx = i                       ; /* Index in the "iolib_moddef[]" table*/
    *maj   = j                       ; /* Index in the "vmos_driver[]" table */

    exit                             : /* Exit label                         */

#ifdef TEST                            /* Begin conditional compilation      */
    procfin("io_chk_driver")         ; /* Message                            */
#endif                                 /* End   conditional compilation      */

    return( r )                      ; /* Return code of the procedure       */

  END_PROC                             /* End of the procedure               */


/*  Procedure io_chk_iod ----------------------------------------------------*/
/*
    Purpose : Checks the an IOD is one allocated by the present module. We
              do that because user may allocate other IOD's, by-passing the
              present module. So, here we check that an IOD has been allocated
              by the present module.

    Author : JYA                                 Last modification : 10/22/2022

...............................................................................

    Calling example : io_chk_iod(iod, typ, moidx, ioidx )

    Input     : iod      INT        IOD to be checked
                typ      INT        IOD type 0:device 1:subchan 2:any

    Output    : moidx    INT        Index in the "iolib_modd[]" table
                ioidx    INT        Index in the "iolib_iod[]" table
                valproc  IWORD      0:Sucess BAD_IOD:Not a "vmk_iolib" IOD

    Notes : - "typ" may be 0:device IOD only 1:subchannel IOD only 2:Any

...............................................................................
*/

static IWORD  PROCEDURE io_chk_iod(INT iod, INT typ, INT *moidx, INT *ioidx)

  BEGIN                                /* Begin of this procedure            */
 
/*  Declaration of local variables of this procedure ........................*/

    IWORD          r                 ; /* Return code of the procedure       */
    INT            maj               ; /* Major number << 16                 */
    IWORD          i                 ; /* Loop counter with MODD's           */
    IWORD          j                 ; /* Loop counter with IOD's            */

/*  Beginning of the code ...................................................*/

#ifdef TEST                            /* Begin conditional compilation      */
    procdeb("io_chk_iod")            ; /* Message                            */
#endif                                 /* End   conditional compilation      */

/*****************************************************************************
 * Step  1 : Initialization of local variables:                              *
 * -------   - We set the "r" local variable, indicating that "iod" is none  *
 *             of ours.                                                      *
 *           - We set the the -1 empty value both "moidx" and "ioidx"        *
 *****************************************************************************/

    r      = BAD_IOD                 ; /* Let'a assume we will fail          */
    *moidx =                           /* In case of error, we return those  */
    *ioidx = -1                      ; /* two -1 empty values                */


/*****************************************************************************
 * Step  2 : The present function returns the "moidx" index. In order to     *
 * -------   retreive it, we search for the major number :                   *
 *           - We extract bits b16..23 of "iod", which are the "maj" major   *
 *             number value                                                  *
 *           - Then we loop with the entries of the "iolib_modd[]" table,    *
 *             which holds the MODD's that are known to the present module,  *
 *             we are happy if we found one MODD that holds in bits b16..23  *
 *             the "maj" major number we are looking for, we stop looping    *
 *****************************************************************************/

    maj = iod AND 0xFF0000           ; /* Extract the major number field     */
                                       /* value, bits b16..23                */
    for (i=0; i LT iolib_nbmodd; i++)  /* We loop on the MODD's of the       */
      if((iolib_modd[i] AND 0xFF000)   /* "iolib_modd[]" table, looking for  */
          EQ maj                    )  /* the "maj" major number in bits     */
        goto step3                   ; /* b16..23 of the MODD                */
                                       /*                                    */
    goto exit                        ; /* We failed, exit with BAD_IOD       */


/*****************************************************************************
 * Step  3 : Retreive the "ioidx" index value, which is the index in the     *
 * -------   "iolib_iod[]" table of IOD's. We simply loop on the entries of  *
 *           this table of INT's, searching for "iod", we stop looping when  *
 *           we have found it.                                               *
 *****************************************************************************/

    step3                            : /* Search for the "iod" value         */
                                       /*                                    */
    for (j = 0; j LT iolib_nbiod; j++) /* We loop on the entries of          */
      if (iolib_iod[j] EQ iod)         /* the "iolib_iod[]" table, until     */
        goto step4                  ;  /* "iod" is found                     */
                                       /*                                    */
    goto exit                        ; /* We failed, exit with BAD_IOD       */


/*****************************************************************************
 * Step  4 : Compare the IOD type against the "typ" awaited type:            *
 * --------  - if "typ" is 0, we are waiting for a device IOD with a 0xFF    *
 *             low-order byte, if this is not 0xFF we have an error          *
 *           - if "typ" is 1, we are waiting for a sub-channel IOD, we exit  *
 *             with BAD_IOD if the lower byte is 0xFF (device IOD)           *
 *           - if "typ" is 2 (oar any other value), we are OK and contine    *
 *****************************************************************************/

    step4                            : /* Check channel or sub-channel IOD   */
    iod &= 0xFF                      ; /* Keep the low-order byte, which is  */
                                       /* the sub-channel number of 0xFF     */
    if (typ EQ 0)                      /* If we are waiting for a device IOD */
      BEGIN                            /* there is no channel number but 0xFF*/
        if (iod NE 0xFF) goto exit   ; /* If channel number is not the 0xFF  */
      END_IF                           /* empty value, exit with BAD_IOD     */
                                       /*                                    */
    else if (typ EQ 1)                 /* If we are waiting for a sub-channel*/
      BEGIN                            /* IOD (sub-channel number not 0xFF)  */
        if (iod EQ 0xFF) goto exit   ; /* We exit with BAD_IOD if channel #  */
      END_ELSE                         /* is the 0xFF empty value/number     */


/*****************************************************************************
 * Step  5 : We are successful, "i" is the index in "iolib_modd[] and "j"    *
 * -------   the index in "iolib_iod[]". We return "i" as "moidx" and "j"    *
 *           as "ioidx".                                                     *
 *****************************************************************************/

    *moidx = i                       ; /* Return the "iolib_moidx" index     */
    *ioidx = j                       ; /* Return the "iolib_iod" index       */
    r      = 0                       ; /* No more error                      */

    exit                             : /* Exit label                         */

#ifdef TEST                            /* Begin conditional compilation      */
    procfin("io_chk_iod")            ; /* Message                            */
#endif                                 /* End   conditional compilation      */

    return( r )                      ; /* Return code of the procedure       */

  END_PROC                             /* End of the procedure               */


