/*****************************************************************************/
/*                                                                           */
/*  Hyperpanel OS                       File "include/tsk_ttylib.h"          */
/*  Interface definitions for "tsk_ttylib.c"                                 */
/*                                                                           */
/*  Author : JYA                        Last modification : 07/16/2023       */
/*                                                                           */
/*  (c) Copyright Hyperpanel Lab, All Rights Reserved                        */
/*  IDDN.FR.001.320001.*.S.P.1990.000.10000                                  */
/*                                                                           */
/*****************************************************************************/

/*  Some explanations ========================================================


         1  Defines
            1.1  UI incoming event codes
            1.2  Output format values of "app_cfg[]" values

         2  Structures and typedefs
         3  Public variables
         4  Prototypes of public procedures
         5  RAM data to be included in space_appcont_xyz.c
*/

#ifndef VMK_UITTY_H                    /* If this file is not already        */
#define VMK_UITTY_H                    /* included, include it.              */


/*
    1    Defines
    _____________


    1.1  UI incoming event codes
    ----------------------------
*/

#define  UI_0                  0       /* Menu option 0                      */
#define  UI_1                  1       /* Menu option 1                      */
#define  UI_2                  2       /* Menu option 2                      */
#define  UI_3                  3       /* Menu option 3                      */
#define  UI_4                  4       /* Menu option 4                      */
#define  UI_5                  5       /* Menu option 5                      */
#define  UI_6                  6       /* Menu option 6                      */
#define  UI_7                  7       /* Menu option 7                      */
#define  UI_8                  8       /* Menu option 8                      */
#define  UI_9                  9       /* Menu option 9                      */
#define  UI_DOT               10       /* Key '.'                            */
#define  UI_ENTER             11       /* Key OK/Return                      */
#define  UI_BACK              12       /* Key OK/Return                      */
#define  UI_ESC               13       /* Key OK/Return                      */
#define  UI_SW1               14       /* Switch 1                           */
#define  UI_SW2               15       /* Switch 2                           */
#define  UI_SW3               16       /* Switch 3                           */
#define  UI_SW4               17       /* Switch 4 (USR)                     */
#define  UI_CAPT              18       /* End of capture event               */
#define  UI_TIME              30       /* Hidden option (timeout)            */
#define  UI_I2C               31       /* I2C job end                        */
#define  UI_LIGHT             32       /* Update lights/LEDs                 */
#define  UI_ERROR             -1       /* Treatment has set an error code    */
#define  UI_INVALID           -2       /* Invalid menu option                */
#define  UI_IGNO              -3       /* Simply ignore                      */


/*  1.2  Output format values of "app_cfg[]" values
    -----------------------------------------------

    Note: The values have to be chosen in order that bit b0 is the
    justification indicator b0:0 -> left  b0:1->right
*/

#define F_DECL                 0       /* Decimal integer left  justified    */
#define F_DECR                 1       /* Decimal integer right justified    */
#define F_ENUML                2       /* Enumerated      left  justified    */
#define F_ENUMR                3       /* Enumerated      right justified    */
#define F_IPL                  4       /* IP address      left  justified    */
#define F_IPR                  5       /* IP address      right justified    */

/*
    2    Structures and typedefs
    ____________________________


    3    Public variables
    _____________________
*/

extern INT      ui_ttyiod            ; /* TTY/ASY IOD for sending/receive    */
extern UCHAR*   ui_sndbuf            ; /* TTY/ASY sending buffer address     */
extern INT      ui_page              ; /* Current displayed page number      */
extern INT      ui_err               ; /* UI treatement error code           */
extern CHAR     ui_bufchr[]          ; /* UI input buffer                    */
extern INT      ui_lgchr             ; /* Count of bytes currently in bufchr */
extern UCHAR*   ui_hexadr            ; /* UI displayed hexadecimal dump add  */

extern const CHAR loc_save[]         ; /* ANSI Esc seq save cursor location  */
extern const CHAR loc_restore[]      ; /* ANSI Esc seq restore cursor loc    */
extern const CHAR loc_set[]          ; /* ANSI Esc seq set cursor location   */
/*
    4    Prototypes of public procedures
    ____________________________________
*/

IWORD pmen                (int                                               );
IWORD ptim                (int                                               );
IWORD pgrp                (int                                               );
IWORD plab                (int                                               );
IWORD pchr                (int                                               );
IWORD puid                (int                                               );
IWORD pbrd                (int                                               );
IWORD pcmd                (int                                               );
IWORD phex                (int                                               );
IWORD perr                (int                                               );
IWORD tcfg                (int                                               );
IWORD icfg                (int                                               );
IWORD tchr                (int                                               );
IWORD scfg                (int                                               );
IWORD sval                (int                                               );
IWORD sint                (int                                               );
IWORD sipa                (int                                               );
IWORD sret                (int                                               );
IWORD scon                (int                                               );

int   roll_cfg            (int,int,int                                       );

void  get_mcu_uid         (CHAR*                                             );

#endif                                 /* End of VMK_UITTY_H                 */

#ifdef SPACE_APPCONT_XYZ_DATA_C

UINT           ui_cfg[UI_NBCFG]       ;
UINT           ui_iod[UI_NBCFG]       ;

const INT      ui_nbcfg = UI_NBCFG    ;

#endif                                 /* End of SPACE_APPCONT_XYZ_C         */


